####
# This script demonstrates how to filter and sort groups using the Tableau
# Server Client.
#
# To run the script, you must have installed Python 3.7 or later.
####


import argparse
import logging
import urllib.parse

import tableauserverclient as TSC


def create_example_group(group_name="Example Group", server=None):
    new_group = TSC.GroupItem(group_name)
    try:
        new_group = server.groups.create(new_group)
        print("Created a new project called: '%s'" % group_name)
        print(new_group)
    except TSC.ServerResponseError:
        print("Group '%s' already existed" % group_name)


def main():
    parser = argparse.ArgumentParser(description="Filter and sort groups.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    # This sample has no additional options, yet. If you add some, please add them here

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        group_name = "SALES NORTHWEST"
        # Try to create a group named "SALES NORTHWEST"
        create_example_group(group_name, server)

        group_name = "SALES ROMANIA"
        # Try to create a group named "SALES ROMANIA"
        create_example_group(group_name, server)

        # URL Encode the name of the group that we want to filter on
        # i.e. turn spaces into plus signs
        filter_group_name = urllib.parse.quote_plus(group_name)
        options = TSC.RequestOptions()
        options.filter.add(
            TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.Equals, filter_group_name)
        )

        filtered_groups, _ = server.groups.get(req_options=options)
        # Result can either be a matching group or an empty list
        if filtered_groups:
            group_name = filtered_groups.pop().name
            print(group_name)
        else:
            error = "No project named '{}' found".format(filter_group_name)
            print(error)

        # Or, try the above with the django style filtering
        try:
            group = server.groups.filter(name=filter_group_name)[0]
        except IndexError:
            print(f"No project named '{filter_group_name}' found")
        else:
            print(group.name)

        options = TSC.RequestOptions()
        options.filter.add(
            TSC.Filter(
                TSC.RequestOptions.Field.Name,
                TSC.RequestOptions.Operator.In,
                ["SALES+NORTHWEST", "SALES+ROMANIA", "this_group"],
            )
        )

        options.sort.add(TSC.Sort(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Direction.Desc))

        matching_groups, pagination_item = server.groups.get(req_options=options)
        print("Filtered groups are:")
        for group in matching_groups:
            print(group.name)

        # or, try the above with the django style filtering.

        groups = ["SALES NORTHWEST", "SALES ROMANIA", "this_group"]
        groups = [urllib.parse.quote_plus(group) for group in groups]
        for group in server.groups.filter(name__in=groups).sort("-name"):
            print(group.name)


if __name__ == "__main__":
    main()



import copy
import datetime
import uuid
import xml.etree.ElementTree as ET
from typing import Callable, Dict, List, Optional, Set

from defusedxml.ElementTree import fromstring

from tableauserverclient.datetime_helpers import parse_datetime
from .connection_item import ConnectionItem
from .exceptions import UnpopulatedPropertyError
from .permissions_item import PermissionsRule
from .property_decorators import (
    property_is_boolean,
    property_is_data_acceleration_config,
)
from .revision_item import RevisionItem
from .tag_item import TagItem
from .view_item import ViewItem


class WorkbookItem(object):
    def __init__(self, project_id: str, name: Optional[str] = None, show_tabs: bool = False) -> None:
        self._connections = None
        self._content_url = None
        self._webpage_url = None
        self._created_at = None
        self._id: Optional[str] = None
        self._initial_tags: set = set()
        self._pdf = None
        self._powerpoint = None
        self._preview_image = None
        self._project_name = None
        self._revisions = None
        self._size = None
        self._updated_at = None
        self._views = None
        self.name = name
        self._description = None
        self.owner_id: Optional[str] = None
        self.project_id = project_id
        self.show_tabs = show_tabs
        self.hidden_views: Optional[List[str]] = None
        self.tags: Set[str] = set()
        self.data_acceleration_config = {
            "acceleration_enabled": None,
            "accelerate_now": None,
            "last_updated_at": None,
            "acceleration_status": None,
        }
        self._permissions = None

        return None

    def __repr__(self):
        return "<WorkbookItem {0} '{1}' contentUrl='{2}' project={3}>".format(
            self._id, self.name, self.content_url, self.project_id
        )

    @property
    def connections(self) -> List[ConnectionItem]:
        if self._connections is None:
            error = "Workbook item must be populated with connections first."
            raise UnpopulatedPropertyError(error)
        return self._connections()

    @property
    def permissions(self) -> List[PermissionsRule]:
        if self._permissions is None:
            error = "Workbook item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._permissions()

    @property
    def content_url(self) -> Optional[str]:
        return self._content_url

    @property
    def webpage_url(self) -> Optional[str]:
        return self._webpage_url

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        return self._created_at

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def powerpoint(self) -> bytes:
        if self._powerpoint is None:
            error = "Workbook item must be populated with its powerpoint first."
            raise UnpopulatedPropertyError(error)
        return self._powerpoint()

    @property
    def pdf(self) -> bytes:
        if self._pdf is None:
            error = "Workbook item must be populated with its pdf first."
            raise UnpopulatedPropertyError(error)
        return self._pdf()

    @property
    def preview_image(self) -> bytes:
        if self._preview_image is None:
            error = "Workbook item must be populated with its preview image first."
            raise UnpopulatedPropertyError(error)
        return self._preview_image()

    @property
    def project_id(self) -> Optional[str]:
        return self._project_id

    @project_id.setter
    def project_id(self, value: str):
        self._project_id = value

    @property
    def project_name(self) -> Optional[str]:
        return self._project_name

    @property
    def show_tabs(self) -> bool:
        return self._show_tabs

    @show_tabs.setter
    @property_is_boolean
    def show_tabs(self, value: bool):
        self._show_tabs = value

    @property
    def size(self):
        return self._size

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        return self._updated_at

    @property
    def views(self) -> List[ViewItem]:
        # Views can be set in an initial workbook response OR by a call
        # to Server. Without getting too fancy, I think we can rely on
        # returning a list from the response, until they call
        # populate_workbook, in which case we bind the fetcher and
        # return a callable.
        if self._views is None:
            error = "Workbook item must be populated with views first."
            raise UnpopulatedPropertyError(error)
        elif callable(self._views):
            # We've called `populate_views` on this model
            return self._views()
        else:
            # We had views included in a WorkbookItem response
            return self._views

    @property
    def data_acceleration_config(self):
        return self._data_acceleration_config

    @data_acceleration_config.setter
    @property_is_data_acceleration_config
    def data_acceleration_config(self, value):
        self._data_acceleration_config = value

    @property
    def revisions(self) -> List[RevisionItem]:
        if self._revisions is None:
            error = "Workbook item must be populated with revisions first."
            raise UnpopulatedPropertyError(error)
        return self._revisions()

    def _set_connections(self, connections):
        self._connections = connections

    def _set_permissions(self, permissions):
        self._permissions = permissions

    def _set_views(self, views: Callable[[], List[ViewItem]]) -> None:
        self._views = views

    def _set_pdf(self, pdf: Callable[[], bytes]) -> None:
        self._pdf = pdf

    def _set_powerpoint(self, pptx: Callable[[], bytes]) -> None:
        self._powerpoint = pptx

    def _set_preview_image(self, preview_image: Callable[[], bytes]) -> None:
        self._preview_image = preview_image

    def _set_revisions(self, revisions):
        self._revisions = revisions

    def _parse_common_tags(self, workbook_xml, ns):
        if not isinstance(workbook_xml, ET.Element):
            workbook_xml = fromstring(workbook_xml).find(".//t:workbook", namespaces=ns)
        if workbook_xml is not None:
            (
                _,
                _,
                _,
                _,
                _,
                description,
                updated_at,
                _,
                show_tabs,
                project_id,
                project_name,
                owner_id,
                _,
                _,
                data_acceleration_config,
            ) = self._parse_element(workbook_xml, ns)

            self._set_values(
                None,
                None,
                None,
                None,
                None,
                description,
                updated_at,
                None,
                show_tabs,
                project_id,
                project_name,
                owner_id,
                None,
                None,
                data_acceleration_config,
            )

        return self

    def _set_values(
        self,
        id,
        name,
        content_url,
        webpage_url,
        created_at,
        description,
        updated_at,
        size,
        show_tabs,
        project_id,
        project_name,
        owner_id,
        tags,
        views,
        data_acceleration_config,
    ):
        if id is not None:
            self._id = id
        if name:
            self.name = name
        if content_url:
            self._content_url = content_url
        if webpage_url:
            self._webpage_url = webpage_url
        if created_at:
            self._created_at = created_at
        if description:
            self._description = description
        if updated_at:
            self._updated_at = updated_at
        if size:
            self._size = size
        if show_tabs:
            self._show_tabs = show_tabs
        if project_id:
            self.project_id = project_id
        if project_name:
            self._project_name = project_name
        if owner_id:
            self.owner_id = owner_id
        if tags:
            self.tags = tags
            self._initial_tags = copy.copy(tags)
        if views:
            self._views = views
        if data_acceleration_config is not None:
            self.data_acceleration_config = data_acceleration_config

    @classmethod
    def from_response(cls, resp: str, ns: Dict[str, str]) -> List["WorkbookItem"]:
        all_workbook_items = list()
        parsed_response = fromstring(resp)
        all_workbook_xml = parsed_response.findall(".//t:workbook", namespaces=ns)
        for workbook_xml in all_workbook_xml:
            (
                id,
                name,
                content_url,
                webpage_url,
                created_at,
                description,
                updated_at,
                size,
                show_tabs,
                project_id,
                project_name,
                owner_id,
                tags,
                views,
                data_acceleration_config,
            ) = cls._parse_element(workbook_xml, ns)

            # workaround for Personal Space workbooks which won't have a project
            if not project_id:
                project_id = uuid.uuid4()

            workbook_item = cls(project_id)
            workbook_item._set_values(
                id,
                name,
                content_url,
                webpage_url,
                created_at,
                description,
                updated_at,
                size,
                show_tabs,
                None,
                project_name,
                owner_id,
                tags,
                views,
                data_acceleration_config,
            )
            all_workbook_items.append(workbook_item)
        return all_workbook_items

    @staticmethod
    def _parse_element(workbook_xml, ns):
        id = workbook_xml.get("id", None)
        name = workbook_xml.get("name", None)
        content_url = workbook_xml.get("contentUrl", None)
        webpage_url = workbook_xml.get("webpageUrl", None)
        created_at = parse_datetime(workbook_xml.get("createdAt", None))
        description = workbook_xml.get("description", None)
        updated_at = parse_datetime(workbook_xml.get("updatedAt", None))

        size = workbook_xml.get("size", None)
        if size:
            size = int(size)

        show_tabs = string_to_bool(workbook_xml.get("showTabs", ""))

        project_id = None
        project_name = None
        project_tag = workbook_xml.find(".//t:project", namespaces=ns)
        if project_tag is not None:
            project_id = project_tag.get("id", None)
            project_name = project_tag.get("name", None)

        owner_id = None
        owner_tag = workbook_xml.find(".//t:owner", namespaces=ns)
        if owner_tag is not None:
            owner_id = owner_tag.get("id", None)

        tags = None
        tags_elem = workbook_xml.find(".//t:tags", namespaces=ns)
        if tags_elem is not None:
            all_tags = TagItem.from_xml_element(tags_elem, ns)
            tags = all_tags

        views = None
        views_elem = workbook_xml.find(".//t:views", namespaces=ns)
        if views_elem is not None:
            views = ViewItem.from_xml_element(views_elem, ns)

        data_acceleration_config = {
            "acceleration_enabled": None,
            "accelerate_now": None,
            "last_updated_at": None,
            "acceleration_status": None,
        }
        data_acceleration_elem = workbook_xml.find(".//t:dataAccelerationConfig", namespaces=ns)
        if data_acceleration_elem is not None:
            data_acceleration_config = parse_data_acceleration_config(data_acceleration_elem)

        return (
            id,
            name,
            content_url,
            webpage_url,
            created_at,
            description,
            updated_at,
            size,
            show_tabs,
            project_id,
            project_name,
            owner_id,
            tags,
            views,
            data_acceleration_config,
        )


def parse_data_acceleration_config(data_acceleration_elem):
    data_acceleration_config = dict()

    acceleration_enabled = data_acceleration_elem.get("accelerationEnabled", None)
    if acceleration_enabled is not None:
        acceleration_enabled = string_to_bool(acceleration_enabled)

    accelerate_now = data_acceleration_elem.get("accelerateNow", None)
    if accelerate_now is not None:
        accelerate_now = string_to_bool(accelerate_now)

    last_updated_at = data_acceleration_elem.get("lastUpdatedAt", None)
    if last_updated_at is not None:
        last_updated_at = parse_datetime(last_updated_at)

    acceleration_status = data_acceleration_elem.get("accelerationStatus", None)

    data_acceleration_config["acceleration_enabled"] = acceleration_enabled
    data_acceleration_config["accelerate_now"] = accelerate_now
    data_acceleration_config["last_updated_at"] = last_updated_at
    data_acceleration_config["acceleration_status"] = acceleration_status
    return data_acceleration_config


# Used to convert string represented boolean to a boolean type
def string_to_bool(s: str) -> bool:
    return s.lower() == "true"



import unittest

import requests_mock

import tableauserverclient as TSC
from ._utils import read_xml_asset

GET_FAVORITES_XML = "favorites_get.xml"
ADD_FAVORITE_WORKBOOK_XML = "favorites_add_workbook.xml"
ADD_FAVORITE_VIEW_XML = "favorites_add_view.xml"
ADD_FAVORITE_DATASOURCE_XML = "favorites_add_datasource.xml"
ADD_FAVORITE_PROJECT_XML = "favorites_add_project.xml"


class FavoritesTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)
        self.server.version = "2.5"

        # Fake signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self.server.favorites.baseurl
        self.user = TSC.UserItem("alice", TSC.UserItem.Roles.Viewer)
        self.user._id = "dd2239f6-ddf1-4107-981a-4cf94e415794"

    def test_get(self) -> None:
        response_xml = read_xml_asset(GET_FAVORITES_XML)
        with requests_mock.mock() as m:
            m.get("{0}/{1}".format(self.baseurl, self.user.id), text=response_xml)
            self.server.favorites.get(self.user)
        self.assertIsNotNone(self.user._favorites)
        self.assertEqual(len(self.user.favorites["workbooks"]), 1)
        self.assertEqual(len(self.user.favorites["views"]), 1)
        self.assertEqual(len(self.user.favorites["projects"]), 1)
        self.assertEqual(len(self.user.favorites["datasources"]), 1)

        workbook = self.user.favorites["workbooks"][0]
        view = self.user.favorites["views"][0]
        datasource = self.user.favorites["datasources"][0]
        project = self.user.favorites["projects"][0]

        self.assertEqual(workbook.id, "6d13b0ca-043d-4d42-8c9d-3f3313ea3a00")
        self.assertEqual(view.id, "d79634e1-6063-4ec9-95ff-50acbf609ff5")
        self.assertEqual(datasource.id, "e76a1461-3b1d-4588-bf1b-17551a879ad9")
        self.assertEqual(project.id, "1d0304cd-3796-429f-b815-7258370b9b74")

    def test_add_favorite_workbook(self) -> None:
        response_xml = read_xml_asset(ADD_FAVORITE_WORKBOOK_XML)
        workbook = TSC.WorkbookItem("")
        workbook._id = "6d13b0ca-043d-4d42-8c9d-3f3313ea3a00"
        workbook.name = "Superstore"
        with requests_mock.mock() as m:
            m.put("{0}/{1}".format(self.baseurl, self.user.id), text=response_xml)
            self.server.favorites.add_favorite_workbook(self.user, workbook)

    def test_add_favorite_view(self) -> None:
        response_xml = read_xml_asset(ADD_FAVORITE_VIEW_XML)
        view = TSC.ViewItem()
        view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
        view._name = "ENDANGERED SAFARI"
        with requests_mock.mock() as m:
            m.put("{0}/{1}".format(self.baseurl, self.user.id), text=response_xml)
            self.server.favorites.add_favorite_view(self.user, view)

    def test_add_favorite_datasource(self) -> None:
        response_xml = read_xml_asset(ADD_FAVORITE_DATASOURCE_XML)
        datasource = TSC.DatasourceItem("ee8c6e70-43b6-11e6-af4f-f7b0d8e20760")
        datasource._id = "e76a1461-3b1d-4588-bf1b-17551a879ad9"
        datasource.name = "SampleDS"
        with requests_mock.mock() as m:
            m.put("{0}/{1}".format(self.baseurl, self.user.id), text=response_xml)
            self.server.favorites.add_favorite_datasource(self.user, datasource)

    def test_add_favorite_project(self) -> None:
        self.server.version = "3.1"
        baseurl = self.server.favorites.baseurl
        response_xml = read_xml_asset(ADD_FAVORITE_PROJECT_XML)
        project = TSC.ProjectItem("Tableau")
        project._id = "1d0304cd-3796-429f-b815-7258370b9b74"
        with requests_mock.mock() as m:
            m.put("{0}/{1}".format(baseurl, self.user.id), text=response_xml)
            self.server.favorites.add_favorite_project(self.user, project)

    def test_delete_favorite_workbook(self) -> None:
        workbook = TSC.WorkbookItem("")
        workbook._id = "6d13b0ca-043d-4d42-8c9d-3f3313ea3a00"
        workbook.name = "Superstore"
        with requests_mock.mock() as m:
            m.delete("{0}/{1}/workbooks/{2}".format(self.baseurl, self.user.id, workbook.id))
            self.server.favorites.delete_favorite_workbook(self.user, workbook)

    def test_delete_favorite_view(self) -> None:
        view = TSC.ViewItem()
        view._id = "d79634e1-6063-4ec9-95ff-50acbf609ff5"
        view._name = "ENDANGERED SAFARI"
        with requests_mock.mock() as m:
            m.delete("{0}/{1}/views/{2}".format(self.baseurl, self.user.id, view.id))
            self.server.favorites.delete_favorite_view(self.user, view)

    def test_delete_favorite_datasource(self) -> None:
        datasource = TSC.DatasourceItem("ee8c6e70-43b6-11e6-af4f-f7b0d8e20760")
        datasource._id = "e76a1461-3b1d-4588-bf1b-17551a879ad9"
        datasource.name = "SampleDS"
        with requests_mock.mock() as m:
            m.delete("{0}/{1}/datasources/{2}".format(self.baseurl, self.user.id, datasource.id))
            self.server.favorites.delete_favorite_datasource(self.user, datasource)

    def test_delete_favorite_project(self) -> None:
        self.server.version = "3.1"
        baseurl = self.server.favorites.baseurl
        project = TSC.ProjectItem("Tableau")
        project._id = "1d0304cd-3796-429f-b815-7258370b9b74"
        with requests_mock.mock() as m:
            m.delete("{0}/{1}/projects/{2}".format(baseurl, self.user.id, project.id))
            self.server.favorites.delete_favorite_project(self.user, project)



import copy
import datetime
import xml.etree.ElementTree as ET
from typing import Dict, List, Optional, Set, Tuple

from defusedxml.ElementTree import fromstring

from tableauserverclient.datetime_helpers import parse_datetime
from .connection_item import ConnectionItem
from .exceptions import UnpopulatedPropertyError
from .permissions_item import PermissionsRule
from .property_decorators import (
    property_not_nullable,
    property_is_boolean,
    property_is_enum,
)
from .revision_item import RevisionItem
from .tag_item import TagItem


class DatasourceItem(object):
    class AskDataEnablement:
        Enabled = "Enabled"
        Disabled = "Disabled"
        SiteDefault = "SiteDefault"

    def __repr__(self):
        return "<Datasource {0} '{1}' ({2} parent={3} >".format(
            self._id,
            self.name,
            self.description or "No Description",
            self.project_id,
        )

    def __init__(self, project_id: str, name: Optional[str] = None) -> None:
        self._ask_data_enablement = None
        self._certified = None
        self._certification_note = None
        self._connections = None
        self._content_url: Optional[str] = None
        self._created_at = None
        self._datasource_type = None
        self._description = None
        self._encrypt_extracts = None
        self._has_extracts = None
        self._id: Optional[str] = None
        self._initial_tags: Set = set()
        self._project_name: Optional[str] = None
        self._revisions = None
        self._updated_at = None
        self._use_remote_query_agent = None
        self._webpage_url = None
        self.description = None
        self.name = name
        self.owner_id: Optional[str] = None
        self.project_id = project_id
        self.tags: Set[str] = set()

        self._permissions = None
        self._data_quality_warnings = None

        return None

    @property
    def ask_data_enablement(self) -> Optional[AskDataEnablement]:
        return self._ask_data_enablement

    @ask_data_enablement.setter
    @property_is_enum(AskDataEnablement)
    def ask_data_enablement(self, value: Optional[AskDataEnablement]):
        self._ask_data_enablement = value

    @property
    def connections(self) -> Optional[List[ConnectionItem]]:
        if self._connections is None:
            error = "Datasource item must be populated with connections first."
            raise UnpopulatedPropertyError(error)
        return self._connections()

    @property
    def permissions(self) -> Optional[List[PermissionsRule]]:
        if self._permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._permissions()

    @property
    def content_url(self) -> Optional[str]:
        return self._content_url

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        return self._created_at

    @property
    def certified(self) -> Optional[bool]:
        return self._certified

    @certified.setter
    @property_not_nullable
    @property_is_boolean
    def certified(self, value: Optional[bool]):
        self._certified = value

    @property
    def certification_note(self) -> Optional[str]:
        return self._certification_note

    @certification_note.setter
    def certification_note(self, value: Optional[str]):
        self._certification_note = value

    @property
    def encrypt_extracts(self):
        return self._encrypt_extracts

    @encrypt_extracts.setter
    @property_is_boolean
    def encrypt_extracts(self, value: Optional[bool]):
        self._encrypt_extracts = value

    @property
    def dqws(self):
        if self._data_quality_warnings is None:
            error = "Project item must be populated with dqws first."
            raise UnpopulatedPropertyError(error)
        return self._data_quality_warnings()

    @property
    def has_extracts(self) -> Optional[bool]:
        return self._has_extracts

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def project_id(self) -> str:
        return self._project_id

    @project_id.setter
    @property_not_nullable
    def project_id(self, value: str):
        self._project_id = value

    @property
    def project_name(self) -> Optional[str]:
        return self._project_name

    @property
    def datasource_type(self) -> Optional[str]:
        return self._datasource_type

    @property
    def description(self) -> Optional[str]:
        return self._description

    @description.setter
    def description(self, value: str):
        self._description = value

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        return self._updated_at

    @property
    def use_remote_query_agent(self) -> Optional[bool]:
        return self._use_remote_query_agent

    @use_remote_query_agent.setter
    @property_is_boolean
    def use_remote_query_agent(self, value: bool):
        self._use_remote_query_agent = value

    @property
    def webpage_url(self) -> Optional[str]:
        return self._webpage_url

    @property
    def revisions(self) -> List[RevisionItem]:
        if self._revisions is None:
            error = "Datasource item must be populated with revisions first."
            raise UnpopulatedPropertyError(error)
        return self._revisions()

    def _set_connections(self, connections):
        self._connections = connections

    def _set_permissions(self, permissions):
        self._permissions = permissions

    def _set_data_quality_warnings(self, dqws):
        self._data_quality_warnings = dqws

    def _set_revisions(self, revisions):
        self._revisions = revisions

    def _parse_common_elements(self, datasource_xml, ns):
        if not isinstance(datasource_xml, ET.Element):
            datasource_xml = fromstring(datasource_xml).find(".//t:datasource", namespaces=ns)
        if datasource_xml is not None:
            (
                ask_data_enablement,
                certified,
                certification_note,
                _,
                _,
                _,
                _,
                encrypt_extracts,
                has_extracts,
                _,
                _,
                owner_id,
                project_id,
                project_name,
                _,
                updated_at,
                use_remote_query_agent,
                webpage_url,
            ) = self._parse_element(datasource_xml, ns)
            self._set_values(
                ask_data_enablement,
                certified,
                certification_note,
                None,
                None,
                None,
                None,
                encrypt_extracts,
                has_extracts,
                None,
                None,
                owner_id,
                project_id,
                project_name,
                None,
                updated_at,
                use_remote_query_agent,
                webpage_url,
            )
        return self

    def _set_values(
        self,
        ask_data_enablement,
        certified,
        certification_note,
        content_url,
        created_at,
        datasource_type,
        description,
        encrypt_extracts,
        has_extracts,
        id_,
        name,
        owner_id,
        project_id,
        project_name,
        tags,
        updated_at,
        use_remote_query_agent,
        webpage_url,
    ):
        if ask_data_enablement is not None:
            self._ask_data_enablement = ask_data_enablement
        if certification_note:
            self.certification_note = certification_note
        self.certified = certified  # Always True/False, not conditional
        if content_url:
            self._content_url = content_url
        if created_at:
            self._created_at = created_at
        if datasource_type:
            self._datasource_type = datasource_type
        if description:
            self.description = description
        if encrypt_extracts is not None:
            self.encrypt_extracts = str(encrypt_extracts).lower() == "true"
        if has_extracts is not None:
            self._has_extracts = str(has_extracts).lower() == "true"
        if id_ is not None:
            self._id = id_
        if name:
            self.name = name
        if owner_id:
            self.owner_id = owner_id
        if project_id:
            self.project_id = project_id
        if project_name:
            self._project_name = project_name
        if tags:
            self.tags = tags
            self._initial_tags = copy.copy(tags)
        if updated_at:
            self._updated_at = updated_at
        if use_remote_query_agent is not None:
            self._use_remote_query_agent = str(use_remote_query_agent).lower() == "true"
        if webpage_url:
            self._webpage_url = webpage_url

    @classmethod
    def from_response(cls, resp: str, ns: Dict) -> List["DatasourceItem"]:
        all_datasource_items = list()
        parsed_response = fromstring(resp)
        all_datasource_xml = parsed_response.findall(".//t:datasource", namespaces=ns)

        for datasource_xml in all_datasource_xml:
            (
                ask_data_enablement,
                certified,
                certification_note,
                content_url,
                created_at,
                datasource_type,
                description,
                encrypt_extracts,
                has_extracts,
                id_,
                name,
                owner_id,
                project_id,
                project_name,
                tags,
                updated_at,
                use_remote_query_agent,
                webpage_url,
            ) = cls._parse_element(datasource_xml, ns)
            datasource_item = cls(project_id)
            datasource_item._set_values(
                ask_data_enablement,
                certified,
                certification_note,
                content_url,
                created_at,
                datasource_type,
                description,
                encrypt_extracts,
                has_extracts,
                id_,
                name,
                owner_id,
                None,
                project_name,
                tags,
                updated_at,
                use_remote_query_agent,
                webpage_url,
            )
            all_datasource_items.append(datasource_item)
        return all_datasource_items

    @staticmethod
    def _parse_element(datasource_xml: ET.Element, ns: Dict) -> Tuple:
        id_ = datasource_xml.get("id", None)
        name = datasource_xml.get("name", None)
        datasource_type = datasource_xml.get("type", None)
        description = datasource_xml.get("description", None)
        content_url = datasource_xml.get("contentUrl", None)
        created_at = parse_datetime(datasource_xml.get("createdAt", None))
        updated_at = parse_datetime(datasource_xml.get("updatedAt", None))
        certification_note = datasource_xml.get("certificationNote", None)
        certified = str(datasource_xml.get("isCertified", None)).lower() == "true"
        certification_note = datasource_xml.get("certificationNote", None)
        certified = str(datasource_xml.get("isCertified", None)).lower() == "true"
        content_url = datasource_xml.get("contentUrl", None)
        created_at = parse_datetime(datasource_xml.get("createdAt", None))
        datasource_type = datasource_xml.get("type", None)
        description = datasource_xml.get("description", None)
        encrypt_extracts = datasource_xml.get("encryptExtracts", None)
        has_extracts = datasource_xml.get("hasExtracts", None)
        id_ = datasource_xml.get("id", None)
        name = datasource_xml.get("name", None)
        updated_at = parse_datetime(datasource_xml.get("updatedAt", None))
        use_remote_query_agent = datasource_xml.get("useRemoteQueryAgent", None)
        webpage_url = datasource_xml.get("webpageUrl", None)

        tags = None
        tags_elem = datasource_xml.find(".//t:tags", namespaces=ns)
        if tags_elem is not None:
            tags = TagItem.from_xml_element(tags_elem, ns)

        project_id = None
        project_name = None
        project_elem = datasource_xml.find(".//t:project", namespaces=ns)
        if project_elem is not None:
            project_id = project_elem.get("id", None)
            project_name = project_elem.get("name", None)

        owner_id = None
        owner_elem = datasource_xml.find(".//t:owner", namespaces=ns)
        if owner_elem is not None:
            owner_id = owner_elem.get("id", None)

        ask_data_enablement = None
        ask_data_elem = datasource_xml.find(".//t:askData", namespaces=ns)
        if ask_data_elem is not None:
            ask_data_enablement = ask_data_elem.get("enablement", None)

        return (
            ask_data_enablement,
            certified,
            certification_note,
            content_url,
            created_at,
            datasource_type,
            description,
            encrypt_extracts,
            has_extracts,
            id_,
            name,
            owner_id,
            project_id,
            project_name,
            tags,
            updated_at,
            use_remote_query_agent,
            webpage_url,
        )



from typing import Callable, List, Optional, TYPE_CHECKING

from defusedxml.ElementTree import fromstring

from .exceptions import UnpopulatedPropertyError
from .property_decorators import property_not_empty, property_is_enum
from .reference_item import ResourceReference
from .user_item import UserItem

if TYPE_CHECKING:
    from tableauserverclient.server import Pager


class GroupItem(object):
    tag_name: str = "group"

    class LicenseMode:
        onLogin: str = "onLogin"
        onSync: str = "onSync"

    def __init__(self, name=None, domain_name=None) -> None:
        self._id: Optional[str] = None
        self._license_mode: Optional[str] = None
        self._minimum_site_role: Optional[str] = None
        self._users: Optional[Callable[..., "Pager"]] = None
        self.name: Optional[str] = name
        self.domain_name: Optional[str] = domain_name

    def __str__(self):
        return "{}({!r})".format(self.__class__.__name__, self.__dict__)

    __repr__ = __str__

    @property
    def domain_name(self) -> Optional[str]:
        return self._domain_name

    @domain_name.setter
    def domain_name(self, value: str) -> None:
        self._domain_name = value

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    @property_not_empty
    def name(self, value: str) -> None:
        self._name = value

    @property
    def license_mode(self) -> Optional[str]:
        return self._license_mode

    @license_mode.setter
    @property_is_enum(LicenseMode)
    def license_mode(self, value: str) -> None:
        self._license_mode = value

    @property
    def minimum_site_role(self) -> Optional[str]:
        return self._minimum_site_role

    @minimum_site_role.setter
    @property_is_enum(UserItem.Roles)
    def minimum_site_role(self, value: str) -> None:
        self._minimum_site_role = value

    @property
    def users(self) -> "Pager":
        if self._users is None:
            error = "Group must be populated with users first."
            raise UnpopulatedPropertyError(error)
        #  Each call to `.users` should create a new pager, this just runs the callable
        return self._users()

    def _set_users(self, users: Callable[..., "Pager"]) -> None:
        self._users = users

    @classmethod
    def from_response(cls, resp, ns) -> List["GroupItem"]:
        all_group_items = list()
        parsed_response = fromstring(resp)
        all_group_xml = parsed_response.findall(".//t:group", namespaces=ns)
        for group_xml in all_group_xml:
            name = group_xml.get("name", None)
            group_item = cls(name)
            group_item._id = group_xml.get("id", None)

            # Domain name is returned in a domain element for some calls
            domain_elem = group_xml.find(".//t:domain", namespaces=ns)
            if domain_elem is not None:
                group_item.domain_name = domain_elem.get("name", None)

            # Import element is returned for both local and AD groups (2020.3+)
            import_elem = group_xml.find(".//t:import", namespaces=ns)
            if import_elem is not None:
                group_item.domain_name = import_elem.get("domainName", None)
                group_item.license_mode = import_elem.get("grantLicenseMode", None)
                group_item.minimum_site_role = import_elem.get("siteRole", None)

            all_group_items.append(group_item)
        return all_group_items

    @staticmethod
    def as_reference(id_: str) -> ResourceReference:
        return ResourceReference(id_, GroupItem.tag_name)



# These two imports must come first
from .request_factory import RequestFactory
from .request_options import (
    CSVRequestOptions,
    ExcelRequestOptions,
    ImageRequestOptions,
    PDFRequestOptions,
    RequestOptions,
)

from .filter import Filter
from .sort import Sort
from ..models import *
from .endpoint import *
from .server import Server
from .pager import Pager
from .exceptions import NotSignedInError
from ..helpers import *



import unittest

import requests_mock

import tableauserverclient as TSC
from ._utils import read_xml_asset, asset

GET_XML = "database_get.xml"
POPULATE_PERMISSIONS_XML = "database_populate_permissions.xml"
UPDATE_XML = "database_update.xml"
GET_DQW_BY_CONTENT = "dqw_by_content_type.xml"


class DatabaseTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)
        # Fake signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"
        self.server.version = "3.5"

        self.baseurl = self.server.databases.baseurl

    def test_get(self):
        response_xml = read_xml_asset(GET_XML)
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_databases, pagination_item = self.server.databases.get()

        self.assertEqual(5, pagination_item.total_available)
        self.assertEqual("5ea59b45-e497-4827-8809-bfe213236f75", all_databases[0].id)
        self.assertEqual("hyper", all_databases[0].connection_type)
        self.assertEqual("hyper_0.hyper", all_databases[0].name)

        self.assertEqual("23591f2c-4802-4d6a-9e28-574a8ea9bc4c", all_databases[1].id)
        self.assertEqual("sqlserver", all_databases[1].connection_type)
        self.assertEqual("testv1", all_databases[1].name)
        self.assertEqual("9324cf6b-ba72-4b8e-b895-ac3f28d2f0e0", all_databases[1].contact_id)
        self.assertEqual(True, all_databases[1].certified)

    def test_update(self):
        response_xml = read_xml_asset(UPDATE_XML)
        with requests_mock.mock() as m:
            m.put(self.baseurl + "/23591f2c-4802-4d6a-9e28-574a8ea9bc4c", text=response_xml)
            single_database = TSC.DatabaseItem("test")
            single_database.contact_id = "9324cf6b-ba72-4b8e-b895-ac3f28d2f0e0"
            single_database._id = "23591f2c-4802-4d6a-9e28-574a8ea9bc4c"
            single_database.certified = True
            single_database.certification_note = "Test"
            single_database = self.server.databases.update(single_database)

        self.assertEqual("23591f2c-4802-4d6a-9e28-574a8ea9bc4c", single_database.id)
        self.assertEqual("9324cf6b-ba72-4b8e-b895-ac3f28d2f0e0", single_database.contact_id)
        self.assertEqual(True, single_database.certified)
        self.assertEqual("Test", single_database.certification_note)

    def test_populate_permissions(self):
        with open(asset(POPULATE_PERMISSIONS_XML), "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "/0448d2ed-590d-4fa0-b272-a2a8a24555b5/permissions", text=response_xml)
            single_database = TSC.DatabaseItem("test")
            single_database._id = "0448d2ed-590d-4fa0-b272-a2a8a24555b5"

            self.server.databases.populate_permissions(single_database)
            permissions = single_database.permissions

            self.assertEqual(permissions[0].grantee.tag_name, "group")
            self.assertEqual(permissions[0].grantee.id, "5e5e1978-71fa-11e4-87dd-7382f5c437af")
            self.assertDictEqual(
                permissions[0].capabilities,
                {
                    TSC.Permission.Capability.ChangePermissions: TSC.Permission.Mode.Deny,
                    TSC.Permission.Capability.Read: TSC.Permission.Mode.Allow,
                },
            )

            self.assertEqual(permissions[1].grantee.tag_name, "user")
            self.assertEqual(permissions[1].grantee.id, "7c37ee24-c4b1-42b6-a154-eaeab7ee330a")
            self.assertDictEqual(
                permissions[1].capabilities,
                {
                    TSC.Permission.Capability.Write: TSC.Permission.Mode.Allow,
                },
            )

    def test_populate_data_quality_warning(self):
        with open(asset(GET_DQW_BY_CONTENT), "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(
                self.server.databases._data_quality_warnings.baseurl + "/94441d26-9a52-4a42-b0fb-3f94792d1aac",
                text=response_xml,
            )
            single_database = TSC.DatabaseItem("test")
            single_database._id = "94441d26-9a52-4a42-b0fb-3f94792d1aac"

            self.server.databases.populate_dqw(single_database)
            dqws = single_database.dqws
            first_dqw = dqws.pop()
            self.assertEqual(first_dqw.id, "c2e0e406-84fb-4f4e-9998-f20dd9306710")
            self.assertEqual(first_dqw.warning_type, "WARNING")
            self.assertEqual(first_dqw.message, "Hello, World!")
            self.assertEqual(first_dqw.owner_id, "eddc8c5f-6af0-40be-b6b0-2c790290a43f")
            self.assertEqual(first_dqw.active, True)
            self.assertEqual(first_dqw.severe, True)
            self.assertEqual(str(first_dqw.created_at), "2021-04-09 18:39:54+00:00")
            self.assertEqual(str(first_dqw.updated_at), "2021-04-09 18:39:54+00:00")

    def test_delete(self):
        with requests_mock.mock() as m:
            m.delete(self.baseurl + "/0448d2ed-590d-4fa0-b272-a2a8a24555b5", status_code=204)
            self.server.databases.delete("0448d2ed-590d-4fa0-b272-a2a8a24555b5")



from defusedxml.ElementTree import fromstring
from typing import Optional


class TableauError(Exception):
    pass


class ServerResponseError(TableauError):
    def __init__(self, code, summary, detail, url=None):
        self.code = code
        self.summary = summary
        self.detail = detail
        self.url = url
        super(ServerResponseError, self).__init__(str(self))

    def __str__(self):
        return "\n\n\t{0}: {1}\n\t\t{2}".format(self.code, self.summary, self.detail)

    @classmethod
    def from_response(cls, resp, ns, url=None):
        # Check elements exist before .text
        parsed_response = fromstring(resp)
        try:
            error_response = cls(
                parsed_response.find("t:error", namespaces=ns).get("code", ""),
                parsed_response.find(".//t:summary", namespaces=ns).text,
                parsed_response.find(".//t:detail", namespaces=ns).text,
                url,
            )
        except Exception as e:
            raise NonXMLResponseError(resp)
        return error_response


class InternalServerError(TableauError):
    def __init__(self, server_response, request_url: Optional[str] = None):
        self.code = server_response.status_code
        self.content = server_response.content
        self.url = request_url or "server"

    def __str__(self):
        return "\n\nInternal error {0} at {1}\n{2}".format(self.code, self.url, self.content)


class MissingRequiredFieldError(TableauError):
    pass


class ServerInfoEndpointNotFoundError(TableauError):
    pass


class EndpointUnavailableError(TableauError):
    pass


class ItemTypeNotAllowed(TableauError):
    pass


class NonXMLResponseError(TableauError):
    pass


class InvalidGraphQLQuery(TableauError):
    pass


class GraphQLError(TableauError):
    def __init__(self, error_payload):
        self.error = error_payload

    def __str__(self):
        from pprint import pformat

        return pformat(self.error)


class JobFailedException(TableauError):
    def __init__(self, job):
        self.notes = job.notes
        self.job = job

    def __str__(self):
        return f"Job {self.job.id} failed with notes {self.notes}"


class JobCancelledException(JobFailedException):
    pass


class FlowRunFailedException(TableauError):
    def __init__(self, flow_run):
        self.background_job_id = flow_run.background_job_id
        self.flow_run = flow_run

    def __str__(self):
        return f"FlowRun {self.flow_run.id} failed with job id {self.background_job_id}"


class FlowRunCancelledException(FlowRunFailedException):
    pass



####
# This script demonstrates how to work with pagination in the .get() method calls, and how to use
# the QuerySet item that is an alternative interface for filtering and sorting these calls.
#
# In Part 1, this script will iterate over every workbook that exists on the server using the
# pagination item to fetch additional pages as needed.
# In Part 2, the script will iterate over the same workbooks with an easy-to-read filter.
#
# While this sample uses workbooks, this same technique will work with any of the .get() methods that return
# a pagination item
####

import argparse
import logging
import os.path

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Demonstrate pagination on the list of workbooks on the server.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-n", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample:
    # No additional options, yet. If you add some, please add them here

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    # Sign in to server
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        you = server.users.get_by_id(server.user_id)
        print(you.name, you.id)
        # 1. Pager: Pager takes a server Endpoint as its first parameter, and a RequestOptions
        # object as the second parameter. The Endpoint defines which type of objects it returns, and
        # RequestOptions defines any restrictions on the objects: filter by name, sort, or select a page
        print("Your server contains the following workbooks:\n")
        count = 0
        # Using a small number here so that you can see it work. Default is 100 and mostly doesn't need to change
        page_options = TSC.RequestOptions(1, 5)
        print("Fetching workbooks in pages of 5")
        for wb in TSC.Pager(server.workbooks, page_options):
            print(wb.name)
            count = count + 1
        print("Total: {}\n".format(count))

        count = 0
        page_options = TSC.RequestOptions(2, 3)
        print("Paging: start at the second page of workbooks, using pagesize = 3")
        for wb in TSC.Pager(server.workbooks, page_options):
            print(wb.name)
            count = count + 1
        print("Truncated Total: {}\n".format(count))

        print("Your id: ", you.name, you.id, "\n")
        count = 0
        filtered_page_options = TSC.RequestOptions(1, 3)
        filter_owner = TSC.Filter("ownerEmail", TSC.RequestOptions.Operator.Equals, "jfitzgerald@tableau.com")
        filtered_page_options.filter.add(filter_owner)
        print("Fetching workbooks again, filtering by owner")
        for wb in TSC.Pager(server.workbooks, filtered_page_options):
            print(wb.name, " -- ", wb.owner_id)
            count = count + 1
        print("Filtered Total: {}\n".format(count))

        # 2. QuerySet offers a fluent interface on top of the RequestOptions object
        print("Fetching workbooks again - this time filtered with QuerySet")
        count = 0
        page = 1
        more = True
        while more:
            queryset = server.workbooks.filter(ownerEmail="jfitzgerald@tableau.com")
            for wb in queryset.paginate(page_number=page, page_size=3):
                print(wb.name, " -- ", wb.owner_id)
                count = count + 1
            more = queryset.total_available > count
            page = page + 1
        print("QuerySet Total: {}".format(count))

        # 3. QuerySet also allows you to iterate over all objects without explicitly paging.
        print("Fetching again - this time without manually paging")
        for i, wb in enumerate(server.workbooks.filter(owner_email="jfitzgerald@tableau.com"), start=1):
            print(wb.name, "--", wb.owner_id)

        print(f"QuerySet Total, implicit paging: {i}")


if __name__ == "__main__":
    main()



import logging

from .endpoint import Endpoint
from .exceptions import MissingRequiredFieldError
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import DatabaseItem, PermissionsRule, ProjectItem, plural_type, Resource
from typing import TYPE_CHECKING, Callable, List, Optional, Sequence, Union

if TYPE_CHECKING:
    from ..server import Server
    from ..request_options import RequestOptions

logger = logging.getLogger(__name__)

# these are the only two items that can hold default permissions for another type
BaseItem = Union[DatabaseItem, ProjectItem]


class _DefaultPermissionsEndpoint(Endpoint):
    """Adds default-permission model to an existing database or project

    Tableau default-permissions model takes an object type in the uri to set the defaults.
    This class is meant to be instantiated inside a parent endpoint which
    has these supported endpoints
    """

    def __init__(self, parent_srv: "Server", owner_baseurl: Callable[[], str]) -> None:
        super(_DefaultPermissionsEndpoint, self).__init__(parent_srv)

        # owner_baseurl is the baseurl of the parent, a project or database.
        # It MUST be a lambda since we don't know the full site URL until we sign in.
        # If populated without, we will get a sign-in error
        self.owner_baseurl = owner_baseurl

    def __str__(self):
        return "<DefaultPermissionsEndpoint {} [Flow, Datasource, Workbook, Lens]>".format(self.owner_baseurl())

    __repr__ = __str__

    def update_default_permissions(
        self, resource: BaseItem, permissions: Sequence[PermissionsRule], content_type: Resource
    ) -> List[PermissionsRule]:
        url = "{0}/{1}/default-permissions/{2}".format(self.owner_baseurl(), resource.id, plural_type(content_type))
        update_req = RequestFactory.Permission.add_req(permissions)
        response = self.put_request(url, update_req)
        permissions = PermissionsRule.from_response(response.content, self.parent_srv.namespace)
        logger.info("Updated default {} permissions for resource {}".format(content_type, resource.id))
        logger.info(permissions)

        return permissions

    def delete_default_permission(self, resource: BaseItem, rule: PermissionsRule, content_type: Resource) -> None:
        for capability, mode in rule.capabilities.items():
            # Made readability better but line is too long, will make this look better
            url = (
                "{baseurl}/{content_id}/default-permissions/"
                "{content_type}/{grantee_type}/{grantee_id}/{cap}/{mode}".format(
                    baseurl=self.owner_baseurl(),
                    content_id=resource.id,
                    content_type=plural_type(content_type),
                    grantee_type=rule.grantee.tag_name + "s",
                    grantee_id=rule.grantee.id,
                    cap=capability,
                    mode=mode,
                )
            )

            logger.debug("Removing {0} permission for capability {1}".format(mode, capability))

            self.delete_request(url)

        logger.info(
            "Deleted permission for {0} {1} item {2}".format(rule.grantee.tag_name, rule.grantee.id, resource.id)
        )

    def populate_default_permissions(self, item: BaseItem, content_type: Resource) -> None:
        if not item.id:
            error = "Server item is missing ID. Item must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        def permission_fetcher() -> List[PermissionsRule]:
            return self._get_default_permissions(item, content_type)

        item._set_default_permissions(permission_fetcher, content_type)
        logger.info("Populated default {0} permissions for item (ID: {1})".format(content_type, item.id))

    def _get_default_permissions(
        self, item: BaseItem, content_type: Resource, req_options: Optional["RequestOptions"] = None
    ) -> List[PermissionsRule]:
        url = "{0}/{1}/default-permissions/{2}".format(self.owner_baseurl(), item.id, plural_type(content_type))
        server_response = self.get_request(url, req_options)
        permissions = PermissionsRule.from_response(server_response.content, self.parent_srv.namespace)
        logger.info({"content_type": content_type, "permissions": permissions})
        return permissions



import time

# Polling for server-side events (such as job completion) uses exponential backoff for the sleep intervals between polls
ASYNC_POLL_MIN_INTERVAL = 0.5
ASYNC_POLL_MAX_INTERVAL = 30
ASYNC_POLL_BACKOFF_FACTOR = 1.4


class ExponentialBackoffTimer:
    def __init__(self, *, timeout=None):
        self.start_time = time.time()
        self.timeout = timeout
        self.current_sleep_interval = ASYNC_POLL_MIN_INTERVAL

    def sleep(self):
        max_sleep_time = ASYNC_POLL_MAX_INTERVAL
        if self.timeout is not None:
            elapsed = time.time() - self.start_time
            if elapsed >= self.timeout:
                raise TimeoutError(f"Timeout after {elapsed} seconds waiting for asynchronous event")
            remaining_time = self.timeout - elapsed
            # Usually, we would sleep for `ASYNC_POLL_MAX_INTERVAL`, but we don't want to sleep over the timeout
            max_sleep_time = min(ASYNC_POLL_MAX_INTERVAL, remaining_time)
            # We want to sleep at least for `ASYNC_POLL_MIN_INTERVAL`. This is important to ensure that, as we get
            # closer to the timeout, we don't accidentally wake up multiple times and hit the server in rapid succession
            # due to waking up to early from the `sleep`.
            max_sleep_time = max(max_sleep_time, ASYNC_POLL_MIN_INTERVAL)

        time.sleep(min(self.current_sleep_interval, max_sleep_time))
        self.current_sleep_interval *= ASYNC_POLL_BACKOFF_FACTOR



from defusedxml.ElementTree import fromstring

from tableauserverclient.datetime_helpers import parse_datetime
from .schedule_item import ScheduleItem
from .target import Target


class TaskItem(object):
    class Type:
        ExtractRefresh = "extractRefresh"
        DataAcceleration = "dataAcceleration"
        RunFlow = "runFlow"

    # This mapping is used to convert task type returned from server
    _TASK_TYPE_MAPPING = {
        "RefreshExtractTask": Type.ExtractRefresh,
        "MaterializeViewsTask": Type.DataAcceleration,
    }

    def __init__(
        self,
        id_,
        task_type,
        priority,
        consecutive_failed_count=0,
        schedule_id=None,
        schedule_item=None,
        last_run_at=None,
        target=None,
    ):
        self.id = id_
        self.task_type = task_type
        self.priority = priority
        self.consecutive_failed_count = consecutive_failed_count
        self.schedule_id = schedule_id
        self.schedule_item = schedule_item
        self.last_run_at = last_run_at
        self.target = target

    def __repr__(self):
        return (
            "<Task#{id} {task_type} pri({priority}) failed({consecutive_failed_count}) schedule_id({"
            "schedule_id}) target({target})>".format(**self.__dict__)
        )

    @classmethod
    def from_response(cls, xml, ns, task_type=Type.ExtractRefresh):
        parsed_response = fromstring(xml)
        all_tasks_xml = parsed_response.findall(".//t:task/t:{}".format(task_type), namespaces=ns)

        all_tasks = (TaskItem._parse_element(x, ns) for x in all_tasks_xml)

        return list(all_tasks)

    @classmethod
    def _parse_element(cls, element, ns):
        schedule_item = None
        target = None
        last_run_at = None
        workbook_element = element.find(".//t:workbook", namespaces=ns)
        datasource_element = element.find(".//t:datasource", namespaces=ns)
        last_run_at_element = element.find(".//t:lastRunAt", namespaces=ns)

        schedule_item_list = ScheduleItem.from_element(element, ns)
        if len(schedule_item_list) >= 1:
            schedule_item = schedule_item_list[0]

        # according to the Tableau Server REST API documentation,
        # there should be only one of workbook or datasource
        if workbook_element is not None:
            workbook_id = workbook_element.get("id", None)
            target = Target(workbook_id, "workbook")
        if datasource_element is not None:
            datasource_id = datasource_element.get("id", None)
            target = Target(datasource_id, "datasource")
        if last_run_at_element is not None:
            last_run_at = parse_datetime(last_run_at_element.text)

        # Server response has different names for task types
        task_type = cls._translate_task_type(element.get("type", None))

        priority = int(element.get("priority", -1))
        consecutive_failed_count = int(element.get("consecutiveFailedCount", 0))
        id_ = element.get("id", None)
        return cls(
            id_,
            task_type,
            priority,
            consecutive_failed_count,
            schedule_item.id,
            schedule_item,
            last_run_at,
            target,
        )

    @staticmethod
    def _translate_task_type(task_type):
        if task_type in TaskItem._TASK_TYPE_MAPPING:
            return TaskItem._TASK_TYPE_MAPPING[task_type]
        else:
            return task_type



from defusedxml.ElementTree import fromstring

from .exceptions import UnpopulatedPropertyError
from .property_decorators import property_not_empty, property_is_boolean


class TableItem(object):
    def __init__(self, name, description=None):
        self._id = None
        self.description = description
        self.name = name

        self._contact_id = None
        self._certified = None
        self._certification_note = None
        self._permissions = None
        self._schema = None

        self._columns = None
        self._data_quality_warnings = None

    @property
    def permissions(self):
        if self._permissions is None:
            error = "Project item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._permissions()

    @property
    def dqws(self):
        if self._data_quality_warnings is None:
            error = "Project item must be populated with dqws first."
            raise UnpopulatedPropertyError(error)
        return self._data_quality_warnings()

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @name.setter
    @property_not_empty
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def certified(self):
        return self._certified

    @certified.setter
    @property_is_boolean
    def certified(self, value):
        self._certified = value

    @property
    def certification_note(self):
        return self._certification_note

    @certification_note.setter
    def certification_note(self, value):
        self._certification_note = value

    @property
    def contact_id(self):
        return self._contact_id

    @contact_id.setter
    def contact_id(self, value):
        self._contact_id = value

    @property
    def schema(self):
        return self._schema

    @property
    def columns(self):
        if self._columns is None:
            error = "Table must be populated with columns first."
            raise UnpopulatedPropertyError(error)
        #  Each call to `.columns` should create a new pager, this just runs the callable
        return self._columns()

    def _set_columns(self, columns):
        self._columns = columns

    def _set_data_quality_warnings(self, dqws):
        self._data_quality_warnings = dqws

    def _set_values(self, table_values):
        if "id" in table_values:
            self._id = table_values["id"]

        if "name" in table_values:
            self._name = table_values["name"]

        if "description" in table_values:
            self._description = table_values["description"]

        if "isCertified" in table_values:
            self._certified = string_to_bool(table_values["isCertified"])

        if "certificationNote" in table_values:
            self._certification_note = table_values["certificationNote"]

        if "isEmbedded" in table_values:
            self._embedded = string_to_bool(table_values["isEmbedded"])

        if "schema" in table_values:
            self._schema = table_values["schema"]

        if "contact" in table_values:
            self._contact_id = table_values["contact"]["id"]

    def _set_permissions(self, permissions):
        self._permissions = permissions

    @classmethod
    def from_response(cls, resp, ns):
        all_table_items = list()
        parsed_response = fromstring(resp)
        all_table_xml = parsed_response.findall(".//t:table", namespaces=ns)

        for table_xml in all_table_xml:
            parsed_table = cls._parse_element(table_xml, ns)
            table_item = cls(parsed_table["name"])
            table_item._set_values(parsed_table)
            all_table_items.append(table_item)
        return all_table_items

    @staticmethod
    def _parse_element(table_xml, ns):
        table_values = table_xml.attrib.copy()

        contact = table_xml.find(".//t:contact", namespaces=ns)
        if contact is not None:
            table_values["contact"] = contact.attrib.copy()

        return table_values


# Used to convert string represented boolean to a boolean type
def string_to_bool(s):
    return s.lower() == "true"



from .property_decorators import property_is_valid_time, property_not_nullable


class IntervalItem(object):
    class Frequency:
        Hourly = "Hourly"
        Daily = "Daily"
        Weekly = "Weekly"
        Monthly = "Monthly"

    class Occurrence:
        Minutes = "minutes"
        Hours = "hours"
        WeekDay = "weekDay"
        MonthDay = "monthDay"

    class Day:
        Sunday = "Sunday"
        Monday = "Monday"
        Tuesday = "Tuesday"
        Wednesday = "Wednesday"
        Thursday = "Thursday"
        Friday = "Friday"
        Saturday = "Saturday"
        LastDay = "LastDay"


class HourlyInterval(object):
    def __init__(self, start_time, end_time, interval_value):
        self.start_time = start_time
        self.end_time = end_time
        self.interval = interval_value

    @property
    def _frequency(self):
        return IntervalItem.Frequency.Hourly

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    @property_is_valid_time
    @property_not_nullable
    def start_time(self, value):
        self._start_time = value

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    @property_is_valid_time
    @property_not_nullable
    def end_time(self, value):
        self._end_time = value

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, interval):
        VALID_INTERVALS = {0.25, 0.5, 1, 2, 4, 6, 8, 12}
        if float(interval) not in VALID_INTERVALS:
            error = "Invalid interval {} not in {}".format(interval, str(VALID_INTERVALS))
            raise ValueError(error)

        self._interval = interval

    def _interval_type_pairs(self):
        # We use fractional hours for the two minute-based intervals.
        # Need to convert to minutes from hours here
        if self.interval in {0.25, 0.5}:
            calculated_interval = int(self.interval * 60)
            interval_type = IntervalItem.Occurrence.Minutes
        else:
            calculated_interval = self.interval
            interval_type = IntervalItem.Occurrence.Hours

        return [(interval_type, str(calculated_interval))]


class DailyInterval(object):
    def __init__(self, start_time, *interval_values):
        self.start_time = start_time
        self.interval = interval_values

    @property
    def _frequency(self):
        return IntervalItem.Frequency.Daily

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    @property_is_valid_time
    @property_not_nullable
    def start_time(self, value):
        self._start_time = value

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, interval):
        self._interval = interval


class WeeklyInterval(object):
    def __init__(self, start_time, *interval_values):
        self.start_time = start_time
        self.interval = interval_values

    @property
    def _frequency(self):
        return IntervalItem.Frequency.Weekly

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    @property_is_valid_time
    @property_not_nullable
    def start_time(self, value):
        self._start_time = value

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, interval_values):
        if not all(hasattr(IntervalItem.Day, day) for day in interval_values):
            raise ValueError("Invalid week day defined " + str(interval_values))

        self._interval = interval_values

    def _interval_type_pairs(self):
        return [(IntervalItem.Occurrence.WeekDay, day) for day in self.interval]


class MonthlyInterval(object):
    def __init__(self, start_time, interval_value):
        self.start_time = start_time
        self.interval = str(interval_value)

    @property
    def _frequency(self):
        return IntervalItem.Frequency.Monthly

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    @property_is_valid_time
    @property_not_nullable
    def start_time(self, value):
        self._start_time = value

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, interval_value):
        error = "Invalid interval value for a monthly frequency: {}.".format(interval_value)

        # This is weird because the value could be a str or an int
        # The only valid str is 'LastDay' so we check that first. If that's not it
        # try to convert it to an int, if that fails because it's an incorrect string
        # like 'badstring' we catch and re-raise. Otherwise we convert to int and check
        # that it's in range 1-31

        if interval_value != "LastDay":
            try:
                if not (1 <= int(interval_value) <= 31):
                    raise ValueError(error)
            except ValueError:
                if interval_value != "LastDay":
                    raise ValueError(error)

        self._interval = str(interval_value)

    def _interval_type_pairs(self):
        return [(IntervalItem.Occurrence.MonthDay, self.interval)]



import logging

from .endpoint import Endpoint, api
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import WebhookItem, PaginationItem

logger = logging.getLogger("tableau.endpoint.webhooks")

from typing import List, Optional, TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from ..server import Server
    from ..request_options import RequestOptions


class Webhooks(Endpoint):
    def __init__(self, parent_srv: "Server") -> None:
        super(Webhooks, self).__init__(parent_srv)

    @property
    def baseurl(self) -> str:
        return "{0}/sites/{1}/webhooks".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    @api(version="3.6")
    def get(self, req_options: Optional["RequestOptions"] = None) -> Tuple[List[WebhookItem], PaginationItem]:
        logger.info("Querying all Webhooks on site")
        url = self.baseurl
        server_response = self.get_request(url, req_options)
        all_webhook_items = WebhookItem.from_response(server_response.content, self.parent_srv.namespace)
        pagination_item = PaginationItem.from_single_page_list(all_webhook_items)
        return all_webhook_items, pagination_item

    @api(version="3.6")
    def get_by_id(self, webhook_id: str) -> WebhookItem:
        if not webhook_id:
            error = "Webhook ID undefined."
            raise ValueError(error)
        logger.info("Querying single webhook (ID: {0})".format(webhook_id))
        url = "{0}/{1}".format(self.baseurl, webhook_id)
        server_response = self.get_request(url)
        return WebhookItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    @api(version="3.6")
    def delete(self, webhook_id: str) -> None:
        if not webhook_id:
            error = "Webhook ID undefined."
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, webhook_id)
        self.delete_request(url)
        logger.info("Deleted single webhook (ID: {0})".format(webhook_id))

    @api(version="3.6")
    def create(self, webhook_item: WebhookItem) -> WebhookItem:
        url = self.baseurl
        create_req = RequestFactory.Webhook.create_req(webhook_item)
        server_response = self.post_request(url, create_req)
        new_webhook = WebhookItem.from_response(server_response.content, self.parent_srv.namespace)[0]

        logger.info("Created new webhook (ID: {0})".format(new_webhook.id))
        return new_webhook

    @api(version="3.6")
    def test(self, webhook_id: str):
        if not webhook_id:
            error = "Webhook ID undefined."
            raise ValueError(error)
        url = "{0}/{1}/test".format(self.baseurl, webhook_id)
        testOutcome = self.get_request(url)
        logger.info("Testing webhook (ID: {0} returned {1})".format(webhook_id, testOutcome))
        return testOutcome



import itertools
from datetime import datetime
from typing import Dict, List, Optional, Type

from defusedxml.ElementTree import fromstring

from tableauserverclient.datetime_helpers import parse_datetime


class FlowRunItem(object):
    def __init__(self) -> None:
        self._id: str = ""
        self._flow_id: Optional[str] = None
        self._status: Optional[str] = None
        self._started_at: Optional[datetime] = None
        self._completed_at: Optional[datetime] = None
        self._progress: Optional[str] = None
        self._background_job_id: Optional[str] = None

    @property
    def id(self) -> str:
        return self._id

    @property
    def flow_id(self) -> Optional[str]:
        return self._flow_id

    @property
    def status(self) -> Optional[str]:
        return self._status

    @property
    def started_at(self) -> Optional[datetime]:
        return self._started_at

    @property
    def completed_at(self) -> Optional[datetime]:
        return self._completed_at

    @property
    def progress(self) -> Optional[str]:
        return self._progress

    @property
    def background_job_id(self) -> Optional[str]:
        return self._background_job_id

    def _set_values(
        self,
        id,
        flow_id,
        status,
        started_at,
        completed_at,
        progress,
        background_job_id,
    ):
        if id is not None:
            self._id = id
        if flow_id is not None:
            self._flow_id = flow_id
        if status is not None:
            self._status = status
        if started_at is not None:
            self._started_at = started_at
        if completed_at is not None:
            self._completed_at = completed_at
        if progress is not None:
            self._progress = progress
        if background_job_id is not None:
            self._background_job_id = background_job_id

    @classmethod
    def from_response(cls: Type["FlowRunItem"], resp: bytes, ns: Optional[Dict]) -> List["FlowRunItem"]:
        all_flowrun_items = list()
        parsed_response = fromstring(resp)
        all_flowrun_xml = itertools.chain(
            parsed_response.findall(".//t:flowRun[@id]", namespaces=ns),
            parsed_response.findall(".//t:flowRuns[@id]", namespaces=ns),
        )

        for flowrun_xml in all_flowrun_xml:
            parsed = cls._parse_element(flowrun_xml, ns)
            flowrun_item = cls()
            flowrun_item._set_values(**parsed)
            all_flowrun_items.append(flowrun_item)
        return all_flowrun_items

    @staticmethod
    def _parse_element(flowrun_xml, ns):
        result = {}
        result["id"] = flowrun_xml.get("id", None)
        result["flow_id"] = flowrun_xml.get("flowId", None)
        result["status"] = flowrun_xml.get("status", None)
        result["started_at"] = parse_datetime(flowrun_xml.get("startedAt", None))
        result["completed_at"] = parse_datetime(flowrun_xml.get("completedAt", None))
        result["progress"] = flowrun_xml.get("progress", None)
        result["background_job_id"] = flowrun_xml.get("backgroundJobId", None)

        return result



####
# This script demonstrates how to use the Tableau Server Client
# to publish a datasource to a Tableau server. It will publish
# a specified datasource to the 'default' project of the provided site.
#
# Some optional arguments are provided to demonstrate async publishing,
# as well as providing connection credentials when publishing. If the
# provided datasource file is over 64MB in size, TSC will automatically
# publish the datasource using the chunking method.
#
# For more information, refer to the documentations:
# (https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_ref_datasources.htm#publish_data_source)
#
# For signing into server, this script uses personal access tokens. For
# more information on personal access tokens, refer to the documentations:
# (https://help.tableau.com/current/server/en-us/security_personal_access_tokens.htm)
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import logging

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Publish a datasource to server.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    parser.add_argument("--file", "-f", required=True, help="filepath to the datasource to publish")
    parser.add_argument("--project", help="Project within which to publish the datasource")
    parser.add_argument("--async", "-a", help="Publishing asynchronously", dest="async_", action="store_true")
    parser.add_argument("--conn-username", help="connection username")
    parser.add_argument("--conn-password", help="connection password")
    parser.add_argument("--conn-embed", help="embed connection password to datasource", action="store_true")
    parser.add_argument("--conn-oauth", help="connection is configured to use oAuth", action="store_true")

    args = parser.parse_args()

    # Ensure that both the connection username and password are provided, or none at all
    if (args.conn_username and not args.conn_password) or (not args.conn_username and args.conn_password):
        parser.error("Both the connection username and password must be provided")

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    # Sign in to server
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        # Empty project_id field will default the publish to the site's default project
        project_id = ""

        # Retrieve the project id, if a project name was passed
        if args.project is not None:
            req_options = TSC.RequestOptions()
            req_options.filter.add(
                TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.Equals, args.project)
            )
            projects = list(TSC.Pager(server.projects, req_options))
            if len(projects) > 1:
                raise ValueError("The project name is not unique")
            project_id = projects[0].id

        # Create a new datasource item to publish
        new_datasource = TSC.DatasourceItem(project_id=project_id)

        # Create a connection_credentials item if connection details are provided
        new_conn_creds = None
        if args.conn_username:
            new_conn_creds = TSC.ConnectionCredentials(
                args.conn_username, args.conn_password, embed=args.conn_embed, oauth=args.conn_oauth
            )

        # Define publish mode - Overwrite, Append, or CreateNew
        publish_mode = TSC.Server.PublishMode.Overwrite

        # Publish datasource
        if args.async_:
            # Async publishing, returns a job_item
            new_job = server.datasources.publish(
                new_datasource, args.file, publish_mode, connection_credentials=new_conn_creds, as_job=True
            )
            print("Datasource published asynchronously. Job ID: {0}".format(new_job.id))
        else:
            # Normal publishing, returns a datasource_item
            new_datasource = server.datasources.publish(
                new_datasource, args.file, publish_mode, connection_credentials=new_conn_creds
            )
            print("Datasource published. Datasource ID: {0}".format(new_datasource.id))


if __name__ == "__main__":
    main()



from ._version import get_versions
from .namespace import NEW_NAMESPACE as DEFAULT_NAMESPACE
from .models import *
from .server import (
    CSVRequestOptions,
    ExcelRequestOptions,
    ImageRequestOptions,
    PDFRequestOptions,
    RequestOptions,
    MissingRequiredFieldError,
    NotSignedInError,
    ServerResponseError,
    Filter,
    Pager,
    Server,
    Sort,
)



import os
import unittest

import requests_mock

import tableauserverclient as TSC
from tableauserverclient.datetime_helpers import parse_datetime
from tableauserverclient.models.task_item import TaskItem

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

GET_XML_NO_WORKBOOK = os.path.join(TEST_ASSET_DIR, "tasks_no_workbook_or_datasource.xml")
GET_XML_WITH_WORKBOOK = os.path.join(TEST_ASSET_DIR, "tasks_with_workbook.xml")
GET_XML_WITH_DATASOURCE = os.path.join(TEST_ASSET_DIR, "tasks_with_datasource.xml")
GET_XML_WITH_WORKBOOK_AND_DATASOURCE = os.path.join(TEST_ASSET_DIR, "tasks_with_workbook_and_datasource.xml")
GET_XML_DATAACCELERATION_TASK = os.path.join(TEST_ASSET_DIR, "tasks_with_dataacceleration_task.xml")
GET_XML_RUN_NOW_RESPONSE = os.path.join(TEST_ASSET_DIR, "tasks_run_now_response.xml")


class TaskTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)
        self.server.version = "3.8"

        # Fake Signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        # default task type is extractRefreshes
        self.baseurl = "{}/{}".format(self.server.tasks.baseurl, "extractRefreshes")

    def test_get_tasks_with_no_workbook(self):
        with open(GET_XML_NO_WORKBOOK, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_tasks, pagination_item = self.server.tasks.get()

        task = all_tasks[0]
        self.assertEqual(None, task.target)

    def test_get_tasks_with_workbook(self):
        with open(GET_XML_WITH_WORKBOOK, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_tasks, pagination_item = self.server.tasks.get()

        task = all_tasks[0]
        self.assertEqual("c7a9327e-1cda-4504-b026-ddb43b976d1d", task.target.id)
        self.assertEqual("workbook", task.target.type)

    def test_get_tasks_with_datasource(self):
        with open(GET_XML_WITH_DATASOURCE, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_tasks, pagination_item = self.server.tasks.get()

        task = all_tasks[0]
        self.assertEqual("c7a9327e-1cda-4504-b026-ddb43b976d1d", task.target.id)
        self.assertEqual("datasource", task.target.type)

    def test_get_tasks_with_workbook_and_datasource(self):
        with open(GET_XML_WITH_WORKBOOK_AND_DATASOURCE, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_tasks, pagination_item = self.server.tasks.get()

        self.assertEqual("workbook", all_tasks[0].target.type)
        self.assertEqual("datasource", all_tasks[1].target.type)
        self.assertEqual("workbook", all_tasks[2].target.type)

    def test_get_task_with_schedule(self):
        with open(GET_XML_WITH_WORKBOOK, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            all_tasks, pagination_item = self.server.tasks.get()

        task = all_tasks[0]
        self.assertEqual("c7a9327e-1cda-4504-b026-ddb43b976d1d", task.target.id)
        self.assertEqual("workbook", task.target.type)
        self.assertEqual("b60b4efd-a6f7-4599-beb3-cb677e7abac1", task.schedule_id)

    def test_delete(self):
        with requests_mock.mock() as m:
            m.delete(self.baseurl + "/c7a9327e-1cda-4504-b026-ddb43b976d1d", status_code=204)
            self.server.tasks.delete("c7a9327e-1cda-4504-b026-ddb43b976d1d")

    def test_delete_missing_id(self):
        self.assertRaises(ValueError, self.server.tasks.delete, "")

    def test_get_materializeviews_tasks(self):
        with open(GET_XML_DATAACCELERATION_TASK, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get("{}/{}".format(self.server.tasks.baseurl, TaskItem.Type.DataAcceleration), text=response_xml)
            all_tasks, pagination_item = self.server.tasks.get(task_type=TaskItem.Type.DataAcceleration)

        task = all_tasks[0]
        self.assertEqual("a462c148-fc40-4670-a8e4-39b7f0c58c7f", task.target.id)
        self.assertEqual("workbook", task.target.type)
        self.assertEqual("b22190b4-6ac2-4eed-9563-4afc03444413", task.schedule_id)
        self.assertEqual(parse_datetime("2019-12-09T22:30:00Z"), task.schedule_item.next_run_at)
        self.assertEqual(parse_datetime("2019-12-09T20:45:04Z"), task.last_run_at)
        self.assertEqual(TSC.TaskItem.Type.DataAcceleration, task.task_type)

    def test_delete_data_acceleration(self):
        with requests_mock.mock() as m:
            m.delete(
                "{}/{}/{}".format(
                    self.server.tasks.baseurl, TaskItem.Type.DataAcceleration, "c9cff7f9-309c-4361-99ff-d4ba8c9f5467"
                ),
                status_code=204,
            )
            self.server.tasks.delete("c9cff7f9-309c-4361-99ff-d4ba8c9f5467", TaskItem.Type.DataAcceleration)

    def test_get_by_id(self):
        with open(GET_XML_WITH_WORKBOOK, "rb") as f:
            response_xml = f.read().decode("utf-8")
        task_id = "f84901ac-72ad-4f9b-a87e-7a3500402ad6"
        with requests_mock.mock() as m:
            m.get("{}/{}".format(self.baseurl, task_id), text=response_xml)
            task = self.server.tasks.get_by_id(task_id)

        self.assertEqual("c7a9327e-1cda-4504-b026-ddb43b976d1d", task.target.id)
        self.assertEqual("workbook", task.target.type)
        self.assertEqual("b60b4efd-a6f7-4599-beb3-cb677e7abac1", task.schedule_id)
        self.assertEqual(TSC.TaskItem.Type.ExtractRefresh, task.task_type)

    def test_run_now(self):
        task_id = "f84901ac-72ad-4f9b-a87e-7a3500402ad6"
        task = TaskItem(task_id, TaskItem.Type.ExtractRefresh, 100)
        with open(GET_XML_RUN_NOW_RESPONSE, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.post("{}/{}/runNow".format(self.baseurl, task_id), text=response_xml)
            job_response_content = self.server.tasks.run(task).decode("utf-8")

        self.assertTrue("7b6b59a8-ac3c-4d1d-2e9e-0b5b4ba8a7b6" in job_response_content)
        self.assertTrue("RefreshExtract" in job_response_content)



####
# This script demonstrates how to use the Tableau Server Client
# to move a workbook from one site to another. It will find
# a workbook that matches a given name, download the workbook,
# and then publish it to the destination site.
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import logging
import shutil
import tempfile

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(
        description="Move one workbook from the"
        "default project of the default site to"
        "the default project of another site."
    )
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    parser.add_argument("--workbook-name", "-w", required=True, help="name of workbook to move")
    parser.add_argument("--destination-site", "-d", required=True, help="name of site to move workbook into")

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    # Step 1: Sign in to both sites on server
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)

    source_server = TSC.Server(args.server)
    dest_server = TSC.Server(args.server)

    with source_server.auth.sign_in(tableau_auth):
        # Step 2: Query workbook to move
        req_option = TSC.RequestOptions()
        req_option.filter.add(
            TSC.Filter(TSC.RequestOptions.Field.Name, TSC.RequestOptions.Operator.Equals, args.workbook_name)
        )
        all_workbooks, pagination_item = source_server.workbooks.get(req_option)

        # Step 3: Download workbook to a temp directory
        if len(all_workbooks) == 0:
            print("No workbook named {} found.".format(args.workbook_name))
        else:
            tmpdir = tempfile.mkdtemp()
            try:
                workbook_path = source_server.workbooks.download(all_workbooks[0].id, tmpdir)

                # Step 4: Check if destination site exists, then sign in to the site
                all_sites, pagination_info = source_server.sites.get()
                found_destination_site = any(
                    (True for site in all_sites if args.destination_site.lower() == site.content_url.lower())
                )
                if not found_destination_site:
                    error = "No site named {} found.".format(args.destination_site)
                    raise LookupError(error)

                tableau_auth.site_id = args.destination_site

                # Signing into another site requires another server object
                # because of the different auth token and site ID.
                with dest_server.auth.sign_in(tableau_auth):
                    # Step 5: Create a new workbook item and publish workbook. Note that
                    # an empty project_id will publish to the 'Default' project.
                    new_workbook = TSC.WorkbookItem(name=args.workbook_name, project_id="")
                    new_workbook = dest_server.workbooks.publish(
                        new_workbook, workbook_path, mode=TSC.Server.PublishMode.Overwrite
                    )
                    print("Successfully moved {0} ({1})".format(new_workbook.name, new_workbook.id))

                # Step 6: Delete workbook from source site and delete temp directory
                source_server.workbooks.delete(all_workbooks[0].id)

            finally:
                shutil.rmtree(tmpdir)


if __name__ == "__main__":
    main()



import unittest
import requests_mock
from pathlib import Path

import tableauserverclient as TSC
from tableauserverclient.datetime_helpers import format_datetime

assets = Path(__file__).parent / "assets"
METRICS_GET = assets / "metrics_get.xml"
METRICS_GET_BY_ID = assets / "metrics_get_by_id.xml"
METRICS_UPDATE = assets / "metrics_update.xml"


class TestMetrics(unittest.TestCase):
    def setUp(self) -> None:
        self.server = TSC.Server("http://test", False)

        # Fake signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"
        self.server.version = "3.9"

        self.baseurl = self.server.metrics.baseurl

    def test_metrics_get(self) -> None:
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=METRICS_GET.read_text())
            all_metrics, pagination_item = self.server.metrics.get()

        self.assertEqual(len(all_metrics), 2)
        self.assertEqual(pagination_item.total_available, 27)
        self.assertEqual(all_metrics[0].id, "6561daa3-20e8-407f-ba09-709b178c0b4a")
        self.assertEqual(all_metrics[0].name, "Example metric")
        self.assertEqual(all_metrics[0].description, "Description of my metric.")
        self.assertEqual(all_metrics[0].webpage_url, "https://test/#/site/site-name/metrics/3")
        self.assertEqual(format_datetime(all_metrics[0].created_at), "2020-01-02T01:02:03Z")
        self.assertEqual(format_datetime(all_metrics[0].updated_at), "2020-01-02T01:02:03Z")
        self.assertEqual(all_metrics[0].suspended, True)
        self.assertEqual(all_metrics[0].project_id, "32e79edb-6cfd-47dc-ad79-e8ec2fbb1d33")
        self.assertEqual(all_metrics[0].project_name, "Default")
        self.assertEqual(all_metrics[0].owner_id, "32e79edb-6cfd-47dc-ad79-e8ec2fbb1d33")
        self.assertEqual(all_metrics[0].view_id, "29dae0cd-1862-4a20-a638-e2c2dfa682d4")
        self.assertEqual(len(all_metrics[0].tags), 0)

        self.assertEqual(all_metrics[1].id, "721760d9-0aa4-4029-87ae-371c956cea07")
        self.assertEqual(all_metrics[1].name, "Another Example metric")
        self.assertEqual(all_metrics[1].description, "Description of another metric.")
        self.assertEqual(all_metrics[1].webpage_url, "https://test/#/site/site-name/metrics/4")
        self.assertEqual(format_datetime(all_metrics[1].created_at), "2020-01-03T01:02:03Z")
        self.assertEqual(format_datetime(all_metrics[1].updated_at), "2020-01-04T01:02:03Z")
        self.assertEqual(all_metrics[1].suspended, False)
        self.assertEqual(all_metrics[1].project_id, "486e0de0-2258-45bd-99cf-b62013e19f4e")
        self.assertEqual(all_metrics[1].project_name, "Assets")
        self.assertEqual(all_metrics[1].owner_id, "1bbbc2b9-847d-443c-9a1f-dbcf112b8814")
        self.assertEqual(all_metrics[1].view_id, "7dbfdb63-a6ca-4723-93ee-4fefc71992d3")
        self.assertEqual(len(all_metrics[1].tags), 2)
        self.assertIn("Test", all_metrics[1].tags)
        self.assertIn("Asset", all_metrics[1].tags)

    def test_metrics_get_by_id(self) -> None:
        luid = "6561daa3-20e8-407f-ba09-709b178c0b4a"
        with requests_mock.mock() as m:
            m.get(f"{self.baseurl}/{luid}", text=METRICS_GET_BY_ID.read_text())
            metric = self.server.metrics.get_by_id(luid)

        self.assertEqual(metric.id, "6561daa3-20e8-407f-ba09-709b178c0b4a")
        self.assertEqual(metric.name, "Example metric")
        self.assertEqual(metric.description, "Description of my metric.")
        self.assertEqual(metric.webpage_url, "https://test/#/site/site-name/metrics/3")
        self.assertEqual(format_datetime(metric.created_at), "2020-01-02T01:02:03Z")
        self.assertEqual(format_datetime(metric.updated_at), "2020-01-02T01:02:03Z")
        self.assertEqual(metric.suspended, True)
        self.assertEqual(metric.project_id, "32e79edb-6cfd-47dc-ad79-e8ec2fbb1d33")
        self.assertEqual(metric.project_name, "Default")
        self.assertEqual(metric.owner_id, "32e79edb-6cfd-47dc-ad79-e8ec2fbb1d33")
        self.assertEqual(metric.view_id, "29dae0cd-1862-4a20-a638-e2c2dfa682d4")
        self.assertEqual(len(metric.tags), 0)

    def test_metrics_delete(self) -> None:
        luid = "6561daa3-20e8-407f-ba09-709b178c0b4a"
        with requests_mock.mock() as m:
            m.delete(f"{self.baseurl}/{luid}")
            self.server.metrics.delete(luid)

    def test_metrics_update(self) -> None:
        luid = "6561daa3-20e8-407f-ba09-709b178c0b4a"
        metric = TSC.MetricItem()
        metric._id = luid

        with requests_mock.mock() as m:
            m.put(f"{self.baseurl}/{luid}", text=METRICS_UPDATE.read_text())
            metric = self.server.metrics.update(metric)

        self.assertEqual(metric.id, "6561daa3-20e8-407f-ba09-709b178c0b4a")
        self.assertEqual(metric.name, "Example metric")
        self.assertEqual(metric.description, "Description of my metric.")
        self.assertEqual(metric.webpage_url, "https://test/#/site/site-name/metrics/3")
        self.assertEqual(format_datetime(metric.created_at), "2020-01-02T01:02:03Z")
        self.assertEqual(format_datetime(metric.updated_at), "2020-01-02T01:02:03Z")
        self.assertEqual(metric.suspended, True)
        self.assertEqual(metric.project_id, "32e79edb-6cfd-47dc-ad79-e8ec2fbb1d33")
        self.assertEqual(metric.project_name, "Default")
        self.assertEqual(metric.owner_id, "32e79edb-6cfd-47dc-ad79-e8ec2fbb1d33")
        self.assertEqual(metric.view_id, "29dae0cd-1862-4a20-a638-e2c2dfa682d4")
        self.assertEqual(len(metric.tags), 0)



import copy
import logging
import urllib.parse

from .endpoint import Endpoint
from .exceptions import EndpointUnavailableError, ServerResponseError
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import TagItem

logger = logging.getLogger("tableau.endpoint.resource_tagger")


class _ResourceTagger(Endpoint):
    # Add new tags to resource
    def _add_tags(self, baseurl, resource_id, tag_set):
        url = "{0}/{1}/tags".format(baseurl, resource_id)
        add_req = RequestFactory.Tag.add_req(tag_set)

        try:
            server_response = self.put_request(url, add_req)
            return TagItem.from_response(server_response.content, self.parent_srv.namespace)
        except ServerResponseError as e:
            if e.code == "404008":
                error = "Adding tags to this resource type is only available with REST API version 2.6 and later."
                raise EndpointUnavailableError(error)
            raise  # Some other error

    # Delete a resource's tag by name
    def _delete_tag(self, baseurl, resource_id, tag_name):
        encoded_tag_name = urllib.parse.quote(tag_name)
        url = "{0}/{1}/tags/{2}".format(baseurl, resource_id, encoded_tag_name)

        try:
            self.delete_request(url)
        except ServerResponseError as e:
            if e.code == "404008":
                error = "Deleting tags from this resource type is only available with REST API version 2.6 and later."
                raise EndpointUnavailableError(error)
            raise  # Some other error

    # Remove and add tags to match the resource item's tag set
    def update_tags(self, baseurl, resource_item):
        if resource_item.tags != resource_item._initial_tags:
            add_set = resource_item.tags - resource_item._initial_tags
            remove_set = resource_item._initial_tags - resource_item.tags
            for tag in remove_set:
                self._delete_tag(baseurl, resource_item.id, tag)
            if add_set:
                resource_item.tags = self._add_tags(baseurl, resource_item.id, add_set)
            resource_item._initial_tags = copy.copy(resource_item.tags)
        logger.info("Updated tags to {0}".format(resource_item.tags))



import copy
from datetime import datetime
from typing import Callable, Iterator, List, Optional, Set

from defusedxml.ElementTree import fromstring

from tableauserverclient.datetime_helpers import parse_datetime
from .exceptions import UnpopulatedPropertyError
from .permissions_item import PermissionsRule
from .tag_item import TagItem


class ViewItem(object):
    def __init__(self) -> None:
        self._content_url: Optional[str] = None
        self._created_at: Optional[datetime] = None
        self._id: Optional[str] = None
        self._image: Optional[Callable[[], bytes]] = None
        self._initial_tags: Set[str] = set()
        self._name: Optional[str] = None
        self._owner_id: Optional[str] = None
        self._preview_image: Optional[Callable[[], bytes]] = None
        self._project_id: Optional[str] = None
        self._pdf: Optional[Callable[[], bytes]] = None
        self._csv: Optional[Callable[[], Iterator[bytes]]] = None
        self._excel: Optional[Callable[[], Iterator[bytes]]] = None
        self._total_views: Optional[int] = None
        self._sheet_type: Optional[str] = None
        self._updated_at: Optional[datetime] = None
        self._workbook_id: Optional[str] = None
        self._permissions: Optional[Callable[[], List[PermissionsRule]]] = None
        self.tags: Set[str] = set()

    def __repr__(self):
        return "<ViewItem {0} '{1}' contentUrl='{2}' project={3}>".format(
            self._id, self.name, self.content_url, self.project_id
        )

    def _set_preview_image(self, preview_image):
        self._preview_image = preview_image

    def _set_image(self, image):
        self._image = image

    def _set_pdf(self, pdf):
        self._pdf = pdf

    def _set_csv(self, csv):
        self._csv = csv

    def _set_excel(self, excel):
        self._excel = excel

    @property
    def content_url(self) -> Optional[str]:
        return self._content_url

    @property
    def created_at(self) -> Optional[datetime]:
        return self._created_at

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def image(self) -> bytes:
        if self._image is None:
            error = "View item must be populated with its png image first."
            raise UnpopulatedPropertyError(error)
        return self._image()

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def owner_id(self) -> Optional[str]:
        return self._owner_id

    @property
    def preview_image(self) -> bytes:
        if self._preview_image is None:
            error = "View item must be populated with its preview image first."
            raise UnpopulatedPropertyError(error)
        return self._preview_image()

    @property
    def project_id(self) -> Optional[str]:
        return self._project_id

    @property
    def pdf(self) -> bytes:
        if self._pdf is None:
            error = "View item must be populated with its pdf first."
            raise UnpopulatedPropertyError(error)
        return self._pdf()

    @property
    def csv(self) -> Iterator[bytes]:
        if self._csv is None:
            error = "View item must be populated with its csv first."
            raise UnpopulatedPropertyError(error)
        return self._csv()

    @property
    def excel(self) -> Iterator[bytes]:
        if self._excel is None:
            error = "View item must be populated with its excel first."
            raise UnpopulatedPropertyError(error)
        return self._excel()

    @property
    def sheet_type(self) -> Optional[str]:
        return self._sheet_type

    @property
    def total_views(self):
        if self._total_views is None:
            error = "Usage statistics must be requested when querying for view."
            raise UnpopulatedPropertyError(error)
        return self._total_views

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @property
    def workbook_id(self) -> Optional[str]:
        return self._workbook_id

    @property
    def permissions(self) -> List[PermissionsRule]:
        if self._permissions is None:
            error = "View item must be populated with permissions first."
            raise UnpopulatedPropertyError(error)
        return self._permissions()

    def _set_permissions(self, permissions: Callable[[], List[PermissionsRule]]) -> None:
        self._permissions = permissions

    @classmethod
    def from_response(cls, resp, ns, workbook_id="") -> List["ViewItem"]:
        return cls.from_xml_element(fromstring(resp), ns, workbook_id)

    @classmethod
    def from_xml_element(cls, parsed_response, ns, workbook_id="") -> List["ViewItem"]:
        all_view_items = list()
        all_view_xml = parsed_response.findall(".//t:view", namespaces=ns)
        for view_xml in all_view_xml:
            view_item = cls()
            usage_elem = view_xml.find(".//t:usage", namespaces=ns)
            workbook_elem = view_xml.find(".//t:workbook", namespaces=ns)
            owner_elem = view_xml.find(".//t:owner", namespaces=ns)
            project_elem = view_xml.find(".//t:project", namespaces=ns)
            tags_elem = view_xml.find(".//t:tags", namespaces=ns)
            view_item._created_at = parse_datetime(view_xml.get("createdAt", None))
            view_item._updated_at = parse_datetime(view_xml.get("updatedAt", None))
            view_item._id = view_xml.get("id", None)
            view_item._name = view_xml.get("name", None)
            view_item._content_url = view_xml.get("contentUrl", None)
            view_item._sheet_type = view_xml.get("sheetType", None)

            if usage_elem is not None:
                total_view = usage_elem.get("totalViewCount", None)
                if total_view:
                    view_item._total_views = int(total_view)

            if owner_elem is not None:
                view_item._owner_id = owner_elem.get("id", None)

            if project_elem is not None:
                view_item._project_id = project_elem.get("id", None)

            if workbook_id:
                view_item._workbook_id = workbook_id
            elif workbook_elem is not None:
                view_item._workbook_id = workbook_elem.get("id", None)

            if tags_elem is not None:
                tags = TagItem.from_xml_element(tags_elem, ns)
                view_item.tags = tags
                view_item._initial_tags = copy.copy(tags)

            all_view_items.append(view_item)
        return all_view_items



import unittest

import requests_mock

import tableauserverclient as TSC
from ._utils import read_xml_asset

GET_XML = "data_acceleration_report.xml"


class DataAccelerationReportTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)

        # Fake signin
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"
        self.server.version = "3.8"

        self.baseurl = self.server.data_acceleration_report.baseurl

    def test_get(self):
        response_xml = read_xml_asset(GET_XML)
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            data_acceleration_report = self.server.data_acceleration_report.get()

        self.assertEqual(2, len(data_acceleration_report.comparison_records))

        self.assertEqual("site-1", data_acceleration_report.comparison_records[0].site)
        self.assertEqual("sheet-1", data_acceleration_report.comparison_records[0].sheet_uri)
        self.assertEqual("0", data_acceleration_report.comparison_records[0].unaccelerated_session_count)
        self.assertEqual("0.0", data_acceleration_report.comparison_records[0].avg_non_accelerated_plt)
        self.assertEqual("1", data_acceleration_report.comparison_records[0].accelerated_session_count)
        self.assertEqual("0.166", data_acceleration_report.comparison_records[0].avg_accelerated_plt)

        self.assertEqual("site-2", data_acceleration_report.comparison_records[1].site)
        self.assertEqual("sheet-2", data_acceleration_report.comparison_records[1].sheet_uri)
        self.assertEqual("2", data_acceleration_report.comparison_records[1].unaccelerated_session_count)
        self.assertEqual("1.29", data_acceleration_report.comparison_records[1].avg_non_accelerated_plt)
        self.assertEqual("3", data_acceleration_report.comparison_records[1].accelerated_session_count)
        self.assertEqual("0.372", data_acceleration_report.comparison_records[1].avg_accelerated_plt)



import os
import unittest

import requests_mock

import tableauserverclient as TSC

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "assets")

GET_XML_PAGE1 = os.path.join(TEST_ASSET_DIR, "workbook_get_page_1.xml")
GET_XML_PAGE2 = os.path.join(TEST_ASSET_DIR, "workbook_get_page_2.xml")
GET_XML_PAGE3 = os.path.join(TEST_ASSET_DIR, "workbook_get_page_3.xml")


class PagerTests(unittest.TestCase):
    def setUp(self):
        self.server = TSC.Server("http://test", False)

        # Fake sign in
        self.server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self.server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self.server.workbooks.baseurl

    def test_pager_with_no_options(self):
        with open(GET_XML_PAGE1, "rb") as f:
            page_1 = f.read().decode("utf-8")
        with open(GET_XML_PAGE2, "rb") as f:
            page_2 = f.read().decode("utf-8")
        with open(GET_XML_PAGE3, "rb") as f:
            page_3 = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            # Register Pager with default request options
            m.get(self.baseurl, text=page_1)

            # Register Pager with some pages
            m.get(self.baseurl + "?pageNumber=1&pageSize=1", text=page_1)
            m.get(self.baseurl + "?pageNumber=2&pageSize=1", text=page_2)
            m.get(self.baseurl + "?pageNumber=3&pageSize=1", text=page_3)

            # No options should get all 3
            workbooks = list(TSC.Pager(self.server.workbooks))
            self.assertTrue(len(workbooks) == 3)

            # Let's check that workbook items aren't duplicates
            wb1, wb2, wb3 = workbooks
            self.assertEqual(wb1.name, "Page1Workbook")
            self.assertEqual(wb2.name, "Page2Workbook")
            self.assertEqual(wb3.name, "Page3Workbook")

    def test_pager_with_options(self):
        with open(GET_XML_PAGE1, "rb") as f:
            page_1 = f.read().decode("utf-8")
        with open(GET_XML_PAGE2, "rb") as f:
            page_2 = f.read().decode("utf-8")
        with open(GET_XML_PAGE3, "rb") as f:
            page_3 = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            # Register Pager with some pages
            m.get(self.baseurl + "?pageNumber=1&pageSize=1", complete_qs=True, text=page_1)
            m.get(self.baseurl + "?pageNumber=2&pageSize=1", complete_qs=True, text=page_2)
            m.get(self.baseurl + "?pageNumber=3&pageSize=1", complete_qs=True, text=page_3)
            m.get(self.baseurl + "?pageNumber=1&pageSize=3", complete_qs=True, text=page_1)

            # Starting on page 2 should get 2 out of 3
            opts = TSC.RequestOptions(2, 1)
            workbooks = list(TSC.Pager(self.server.workbooks, opts))
            self.assertTrue(len(workbooks) == 2)

            # Check that the workbooks are the 2 we think they should be
            wb2, wb3 = workbooks
            self.assertEqual(wb2.name, "Page2Workbook")
            self.assertEqual(wb3.name, "Page3Workbook")

            # Starting on 1 with pagesize of 3 should get all 3
            opts = TSC.RequestOptions(1, 3)
            workbooks = list(TSC.Pager(self.server.workbooks, opts))
            self.assertTrue(len(workbooks) == 3)
            wb1, wb2, wb3 = workbooks
            self.assertEqual(wb1.name, "Page1Workbook")
            self.assertEqual(wb2.name, "Page2Workbook")
            self.assertEqual(wb3.name, "Page3Workbook")

            # Starting on 3 with pagesize of 1 should get the last item
            opts = TSC.RequestOptions(3, 1)
            workbooks = list(TSC.Pager(self.server.workbooks, opts))
            self.assertTrue(len(workbooks) == 1)
            # Should have the last workbook
            wb3 = workbooks.pop()
            self.assertEqual(wb3.name, "Page3Workbook")



import cgi
import copy
import io
import logging
import os
from contextlib import closing
from pathlib import Path
from typing import Iterable, List, Optional, TYPE_CHECKING, Tuple, Union

from .dqw_endpoint import _DataQualityWarningEndpoint
from .endpoint import QuerysetEndpoint, api
from .exceptions import InternalServerError, MissingRequiredFieldError
from .permissions_endpoint import _PermissionsEndpoint
from .resource_tagger import _ResourceTagger
from tableauserverclient.models import FlowItem, PaginationItem, ConnectionItem, JobItem
from tableauserverclient.server import RequestFactory
from tableauserverclient.filesys_helpers import (
    to_filename,
    make_download_path,
    get_file_type,
    get_file_object_size,
)

io_types_r = (io.BytesIO, io.BufferedReader)
io_types_w = (io.BytesIO, io.BufferedWriter)

io_types_r = (io.BytesIO, io.BufferedReader)
io_types_w = (io.BytesIO, io.BufferedWriter)

# The maximum size of a file that can be published in a single request is 64MB
FILESIZE_LIMIT = 1024 * 1024 * 64  # 64MB

ALLOWED_FILE_EXTENSIONS = ["tfl", "tflx"]

logger = logging.getLogger("tableau.endpoint.flows")

if TYPE_CHECKING:
    from .. import DQWItem
    from ..request_options import RequestOptions
    from ...models.permissions_item import PermissionsRule
    from .schedules_endpoint import AddResponse


FilePath = Union[str, os.PathLike]
FileObjectR = Union[io.BufferedReader, io.BytesIO]
FileObjectW = Union[io.BufferedWriter, io.BytesIO]
PathOrFileR = Union[FilePath, FileObjectR]
PathOrFileW = Union[FilePath, FileObjectW]


class Flows(QuerysetEndpoint):
    def __init__(self, parent_srv):
        super(Flows, self).__init__(parent_srv)
        self._resource_tagger = _ResourceTagger(parent_srv)
        self._permissions = _PermissionsEndpoint(parent_srv, lambda: self.baseurl)
        self._data_quality_warnings = _DataQualityWarningEndpoint(self.parent_srv, "flow")

    @property
    def baseurl(self) -> str:
        return "{0}/sites/{1}/flows".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    # Get all flows
    @api(version="3.3")
    def get(self, req_options: Optional["RequestOptions"] = None) -> Tuple[List[FlowItem], PaginationItem]:
        logger.info("Querying all flows on site")
        url = self.baseurl
        server_response = self.get_request(url, req_options)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_flow_items = FlowItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_flow_items, pagination_item

    # Get 1 flow by id
    @api(version="3.3")
    def get_by_id(self, flow_id: str) -> FlowItem:
        if not flow_id:
            error = "Flow ID undefined."
            raise ValueError(error)
        logger.info("Querying single flow (ID: {0})".format(flow_id))
        url = "{0}/{1}".format(self.baseurl, flow_id)
        server_response = self.get_request(url)
        return FlowItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    # Populate flow item's connections
    @api(version="3.3")
    def populate_connections(self, flow_item: FlowItem) -> None:
        if not flow_item.id:
            error = "Flow item missing ID. Flow must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        def connections_fetcher():
            return self._get_flow_connections(flow_item)

        flow_item._set_connections(connections_fetcher)
        logger.info("Populated connections for flow (ID: {0})".format(flow_item.id))

    def _get_flow_connections(self, flow_item, req_options: Optional["RequestOptions"] = None) -> List[ConnectionItem]:
        url = "{0}/{1}/connections".format(self.baseurl, flow_item.id)
        server_response = self.get_request(url, req_options)
        connections = ConnectionItem.from_response(server_response.content, self.parent_srv.namespace)
        return connections

    # Delete 1 flow by id
    @api(version="3.3")
    def delete(self, flow_id: str) -> None:
        if not flow_id:
            error = "Flow ID undefined."
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, flow_id)
        self.delete_request(url)
        logger.info("Deleted single flow (ID: {0})".format(flow_id))

    # Download 1 flow by id
    @api(version="3.3")
    def download(self, flow_id: str, filepath: Optional[PathOrFileW] = None) -> PathOrFileW:
        if not flow_id:
            error = "Flow ID undefined."
            raise ValueError(error)
        url = "{0}/{1}/content".format(self.baseurl, flow_id)

        with closing(self.get_request(url, parameters={"stream": True})) as server_response:
            _, params = cgi.parse_header(server_response.headers["Content-Disposition"])
            if isinstance(filepath, io_types_w):
                for chunk in server_response.iter_content(1024):  # 1KB
                    filepath.write(chunk)
                return_path = filepath
            else:
                filename = to_filename(os.path.basename(params["filename"]))
                download_path = make_download_path(filepath, filename)
                with open(download_path, "wb") as f:
                    for chunk in server_response.iter_content(1024):  # 1KB
                        f.write(chunk)
                return_path = os.path.abspath(download_path)

        logger.info("Downloaded flow to {0} (ID: {1})".format(return_path, flow_id))
        return return_path

    # Update flow
    @api(version="3.3")
    def update(self, flow_item: FlowItem) -> FlowItem:
        if not flow_item.id:
            error = "Flow item missing ID. Flow must be retrieved from server first."
            raise MissingRequiredFieldError(error)

        self._resource_tagger.update_tags(self.baseurl, flow_item)

        # Update the flow itself
        url = "{0}/{1}".format(self.baseurl, flow_item.id)
        update_req = RequestFactory.Flow.update_req(flow_item)
        server_response = self.put_request(url, update_req)
        logger.info("Updated flow item (ID: {0})".format(flow_item.id))
        updated_flow = copy.copy(flow_item)
        return updated_flow._parse_common_elements(server_response.content, self.parent_srv.namespace)

    # Update flow connections
    @api(version="3.3")
    def update_connection(self, flow_item: FlowItem, connection_item: ConnectionItem) -> ConnectionItem:
        url = "{0}/{1}/connections/{2}".format(self.baseurl, flow_item.id, connection_item.id)

        update_req = RequestFactory.Connection.update_req(connection_item)
        server_response = self.put_request(url, update_req)
        connection = ConnectionItem.from_response(server_response.content, self.parent_srv.namespace)[0]

        logger.info("Updated flow item (ID: {0} & connection item {1}".format(flow_item.id, connection_item.id))
        return connection

    @api(version="3.3")
    def refresh(self, flow_item: FlowItem) -> JobItem:
        url = "{0}/{1}/run".format(self.baseurl, flow_item.id)
        empty_req = RequestFactory.Empty.empty_req()
        server_response = self.post_request(url, empty_req)
        new_job = JobItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        return new_job

    # Publish flow
    @api(version="3.3")
    def publish(
        self, flow_item: FlowItem, file: PathOrFileR, mode: str, connections: Optional[List[ConnectionItem]] = None
    ) -> FlowItem:
        if not mode or not hasattr(self.parent_srv.PublishMode, mode):
            error = "Invalid mode defined."
            raise ValueError(error)

        if isinstance(file, (str, os.PathLike)):
            if not os.path.isfile(file):
                error = "File path does not lead to an existing file."
                raise IOError(error)

            filename = os.path.basename(file)
            file_extension = os.path.splitext(filename)[1][1:]
            file_size = os.path.getsize(file)

            # If name is not defined, grab the name from the file to publish
            if not flow_item.name:
                flow_item.name = os.path.splitext(filename)[0]
            if file_extension not in ALLOWED_FILE_EXTENSIONS:
                error = "Only {} files can be published as flows.".format(", ".join(ALLOWED_FILE_EXTENSIONS))
                raise ValueError(error)

        elif isinstance(file, io_types_r):
            if not flow_item.name:
                error = "Flow item must have a name when passing a file object"
                raise ValueError(error)

            file_type = get_file_type(file)
            if file_type == "zip":
                file_extension = "tflx"
            elif file_type == "xml":
                file_extension = "tfl"
            else:
                error = "Unsupported file type {}!".format(file_type)
                raise ValueError(error)

            # Generate filename for file object.
            # This is needed when publishing the flow in a single request
            filename = "{}.{}".format(flow_item.name, file_extension)
            file_size = get_file_object_size(file)

        else:
            raise TypeError("file should be a filepath or file object.")

        # Construct the url with the defined mode
        url = "{0}?flowType={1}".format(self.baseurl, file_extension)
        if mode == self.parent_srv.PublishMode.Overwrite or mode == self.parent_srv.PublishMode.Append:
            url += "&{0}=true".format(mode.lower())

        # Determine if chunking is required (64MB is the limit for single upload method)
        if file_size >= FILESIZE_LIMIT:
            logger.info("Publishing {0} to server with chunking method (flow over 64MB)".format(filename))
            upload_session_id = self.parent_srv.fileuploads.upload(file)
            url = "{0}&uploadSessionId={1}".format(url, upload_session_id)
            xml_request, content_type = RequestFactory.Flow.publish_req_chunked(flow_item, connections)
        else:
            logger.info("Publishing {0} to server".format(filename))

            if isinstance(file, (str, Path)):
                with open(file, "rb") as f:
                    file_contents = f.read()

            elif isinstance(file, io_types_r):
                file_contents = file.read()

            else:
                raise TypeError("file should be a filepath or file object.")

            xml_request, content_type = RequestFactory.Flow.publish_req(flow_item, filename, file_contents, connections)

        # Send the publishing request to server
        try:
            server_response = self.post_request(url, xml_request, content_type)
        except InternalServerError as err:
            if err.code == 504:
                err.content = "Timeout error while publishing. Please use asynchronous publishing to avoid timeouts."
            raise err
        else:
            new_flow = FlowItem.from_response(server_response.content, self.parent_srv.namespace)[0]
            logger.info("Published {0} (ID: {1})".format(filename, new_flow.id))
            return new_flow

    @api(version="3.3")
    def populate_permissions(self, item: FlowItem) -> None:
        self._permissions.populate(item)

    @api(version="3.3")
    def update_permission(self, item, permission_item):
        import warnings

        warnings.warn(
            "Server.flows.update_permission is deprecated, " "please use Server.flows.update_permissions instead.",
            DeprecationWarning,
        )
        self._permissions.update(item, permission_item)

    @api(version="3.3")
    def update_permissions(self, item: FlowItem, permission_item: Iterable["PermissionsRule"]) -> None:
        self._permissions.update(item, permission_item)

    @api(version="3.3")
    def delete_permission(self, item: FlowItem, capability_item: "PermissionsRule") -> None:
        self._permissions.delete(item, capability_item)

    @api(version="3.5")
    def populate_dqw(self, item: FlowItem) -> None:
        self._data_quality_warnings.populate(item)

    @api(version="3.5")
    def update_dqw(self, item: FlowItem, warning: "DQWItem") -> None:
        return self._data_quality_warnings.update(item, warning)

    @api(version="3.5")
    def add_dqw(self, item: FlowItem, warning: "DQWItem") -> None:
        return self._data_quality_warnings.add(item, warning)

    @api(version="3.5")
    def delete_dqw(self, item: FlowItem) -> None:
        self._data_quality_warnings.clear(item)

    # a convenience method
    @api(version="3.3")
    def schedule_flow_run(
        self, schedule_id: str, item: FlowItem
    ) -> List["AddResponse"]:  # actually should return a task
        return self.parent_srv.schedules.add_to_schedule(schedule_id, flow=item)



import logging

from .endpoint import Endpoint, api
from .exceptions import MissingRequiredFieldError
from tableauserverclient.models import TaskItem, PaginationItem
from tableauserverclient.server import RequestFactory

logger = logging.getLogger("tableau.endpoint.tasks")


class Tasks(Endpoint):
    @property
    def baseurl(self):
        return "{0}/sites/{1}/tasks".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    def __normalize_task_type(self, task_type):
        """
        The word for extract refresh used in API URL is "extractRefreshes".
        It is different than the tag "extractRefresh" used in the request body.
        """
        if task_type == TaskItem.Type.ExtractRefresh:
            return "{}es".format(task_type)
        else:
            return task_type

    @api(version="2.6")
    def get(self, req_options=None, task_type=TaskItem.Type.ExtractRefresh):
        if task_type == TaskItem.Type.DataAcceleration:
            self.parent_srv.assert_at_least_version("3.8", "Data Acceleration Tasks")

        logger.info("Querying all {} tasks for the site".format(task_type))

        url = "{0}/{1}".format(self.baseurl, self.__normalize_task_type(task_type))
        server_response = self.get_request(url, req_options)

        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_tasks = TaskItem.from_response(server_response.content, self.parent_srv.namespace, task_type)
        return all_tasks, pagination_item

    @api(version="2.6")
    def get_by_id(self, task_id):
        if not task_id:
            error = "No Task ID provided"
            raise ValueError(error)
        logger.info("Querying a single task by id ({})".format(task_id))
        url = "{}/{}/{}".format(
            self.baseurl,
            self.__normalize_task_type(TaskItem.Type.ExtractRefresh),
            task_id,
        )
        server_response = self.get_request(url)
        return TaskItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    @api(version="2.6")
    def run(self, task_item):
        if not task_item.id:
            error = "Task item missing ID."
            raise MissingRequiredFieldError(error)

        url = "{0}/{1}/{2}/runNow".format(
            self.baseurl,
            self.__normalize_task_type(TaskItem.Type.ExtractRefresh),
            task_item.id,
        )
        run_req = RequestFactory.Task.run_req(task_item)
        server_response = self.post_request(url, run_req)
        return server_response.content  # Todo add typing

    # Delete 1 task by id
    @api(version="3.6")
    def delete(self, task_id, task_type=TaskItem.Type.ExtractRefresh):
        if task_type == TaskItem.Type.DataAcceleration:
            self.parent_srv.assert_at_least_version("3.8", "Data Acceleration Tasks")

        if not task_id:
            error = "No Task ID provided"
            raise ValueError(error)
        url = "{0}/{1}/{2}".format(self.baseurl, self.__normalize_task_type(task_type), task_id)
        self.delete_request(url)
        logger.info("Deleted single task (ID: {0})".format(task_id))



import logging
import unittest
from unittest.mock import *
from typing import List
import io

import pytest

import tableauserverclient as TSC


class UserModelTests(unittest.TestCase):
    def test_invalid_auth_setting(self):
        user = TSC.UserItem("me", TSC.UserItem.Roles.Publisher)
        with self.assertRaises(ValueError):
            user.auth_setting = "Hello"

    def test_invalid_site_role(self):
        user = TSC.UserItem("me", TSC.UserItem.Roles.Publisher)
        with self.assertRaises(ValueError):
            user.site_role = "Hello"


class UserDataTest(unittest.TestCase):
    logger = logging.getLogger("UserDataTest")

    role_inputs = [
        ["creator", "system", "yes", "SiteAdministrator"],
        ["None", "system", "no", "SiteAdministrator"],
        ["explorer", "SysTEm", "no", "SiteAdministrator"],
        ["creator", "site", "yes", "SiteAdministratorCreator"],
        ["explorer", "site", "yes", "SiteAdministratorExplorer"],
        ["creator", "SITE", "no", "SiteAdministratorCreator"],
        ["creator", "none", "yes", "Creator"],
        ["explorer", "none", "yes", "ExplorerCanPublish"],
        ["viewer", "None", "no", "Viewer"],
        ["explorer", "no", "yes", "ExplorerCanPublish"],
        ["EXPLORER", "noNO", "yes", "ExplorerCanPublish"],
        ["explorer", "no", "no", "Explorer"],
        ["unlicensed", "none", "no", "Unlicensed"],
        ["Chef", "none", "yes", "Unlicensed"],
        ["yes", "yes", "yes", "Unlicensed"],
    ]

    valid_import_content = [
        "username, pword, fname, creator, site, yes, email",
        "username, pword, fname, explorer, none, no, email",
        "",
        "u",
        "p",
    ]

    valid_username_content = ["jfitzgerald@tableau.com"]

    usernames = [
        "valid",
        "valid@email.com",
        "domain/valid",
        "domain/valid@tmail.com",
        "va!@#$%^&*()lid",
        "in@v@lid",
        "in valid",
        "",
    ]

    def test_validate_usernames(self):
        TSC.UserItem.validate_username_or_throw(UserDataTest.usernames[0])
        TSC.UserItem.validate_username_or_throw(UserDataTest.usernames[1])
        TSC.UserItem.validate_username_or_throw(UserDataTest.usernames[2])
        TSC.UserItem.validate_username_or_throw(UserDataTest.usernames[3])
        TSC.UserItem.validate_username_or_throw(UserDataTest.usernames[4])
        with self.assertRaises(AttributeError):
            TSC.UserItem.validate_username_or_throw(UserDataTest.usernames[5])
        with self.assertRaises(AttributeError):
            TSC.UserItem.validate_username_or_throw(UserDataTest.usernames[6])

    def test_evaluate_role(self):
        for line in UserDataTest.role_inputs:
            actual = TSC.UserItem.CSVImport._evaluate_site_role(line[0], line[1], line[2])
            assert actual == line[3], line + [actual]

    def test_get_user_detail_empty_line(self):
        test_line = ""
        test_user = TSC.UserItem.CSVImport.create_user_from_line(test_line)
        assert test_user is None

    def test_get_user_detail_standard(self):
        test_line = "username, pword, fname, license, admin, pub, email"
        test_user: TSC.UserItem = TSC.UserItem.CSVImport.create_user_from_line(test_line)
        assert test_user.name == "username", test_user.name
        assert test_user.fullname == "fname", test_user.fullname
        assert test_user.site_role == "Unlicensed", test_user.site_role
        assert test_user.email == "email", test_user.email

    def test_get_user_details_only_username(self):
        test_line = "username"
        test_user: TSC.UserItem = TSC.UserItem.CSVImport.create_user_from_line(test_line)

    def test_populate_user_details_only_some(self):
        values = "username, , , creator, admin"
        user = TSC.UserItem.CSVImport.create_user_from_line(values)
        assert user.name == "username"

    def test_validate_user_detail_standard(self):
        test_line = "username, pword, fname, creator, site, 1, email"
        TSC.UserItem.CSVImport._validate_import_line_or_throw(test_line, UserDataTest.logger)
        TSC.UserItem.CSVImport.create_user_from_line(test_line)

    # for file handling
    def _mock_file_content(self, content: List[str]) -> io.TextIOWrapper:
        # the empty string represents EOF
        # the tests run through the file twice, first to validate then to fetch
        mock = MagicMock(io.TextIOWrapper)
        content.append("")  # EOF
        mock.readline.side_effect = content
        mock.name = "file-mock"
        return mock

    def test_validate_import_file(self):
        test_data = self._mock_file_content(UserDataTest.valid_import_content)
        valid, invalid = TSC.UserItem.CSVImport.validate_file_for_import(test_data, UserDataTest.logger)
        assert valid == 2, "Expected two lines to be parsed, got {}".format(valid)
        assert invalid == [], "Expected no failures, got {}".format(invalid)

    def test_validate_usernames_file(self):
        test_data = self._mock_file_content(UserDataTest.usernames)
        valid, invalid = TSC.UserItem.CSVImport.validate_file_for_import(test_data, UserDataTest.logger)
        assert valid == 5, "Exactly 5 of the lines were valid, counted {}".format(valid + invalid)



import logging

from defusedxml.ElementTree import fromstring

from .endpoint import Endpoint, api
from .exceptions import ServerResponseError
from ..request_factory import RequestFactory

logger = logging.getLogger("tableau.endpoint.auth")


class Auth(Endpoint):
    class contextmgr(object):
        def __init__(self, callback):
            self._callback = callback

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._callback()

    @property
    def baseurl(self):
        return "{0}/auth".format(self.parent_srv.baseurl)

    @api(version="2.0")
    def sign_in(self, auth_req):
        url = "{0}/{1}".format(self.baseurl, "signin")
        signin_req = RequestFactory.Auth.signin_req(auth_req)
        server_response = self.parent_srv.session.post(
            url, data=signin_req, **self.parent_srv.http_options, allow_redirects=False
        )
        # manually handle a redirect so that we send the correct POST request instead of GET
        # this will make e.g http://online.tableau.com work to redirect to http://east.online.tableau.com
        if server_response.status_code == 301:
            server_response = self.parent_srv.session.post(
                server_response.headers["Location"],
                data=signin_req,
                **self.parent_srv.http_options,
                allow_redirects=False,
            )
        self.parent_srv._namespace.detect(server_response.content)
        self._check_status(server_response, url)
        parsed_response = fromstring(server_response.content)
        site_id = parsed_response.find(".//t:site", namespaces=self.parent_srv.namespace).get("id", None)
        user_id = parsed_response.find(".//t:user", namespaces=self.parent_srv.namespace).get("id", None)
        auth_token = parsed_response.find("t:credentials", namespaces=self.parent_srv.namespace).get("token", None)
        self.parent_srv._set_auth(site_id, user_id, auth_token)
        logger.info("Signed into {0} as user with id {1}".format(self.parent_srv.server_address, user_id))
        return Auth.contextmgr(self.sign_out)

    @api(version="3.6")
    def sign_in_with_personal_access_token(self, auth_req):
        # We use the same request that username/password login uses.
        return self.sign_in(auth_req)

    @api(version="2.0")
    def sign_out(self):
        url = "{0}/{1}".format(self.baseurl, "signout")
        # If there are no auth tokens you're already signed out. No-op
        if not self.parent_srv.is_signed_in():
            return
        self.post_request(url, "")
        self.parent_srv._clear_auth()
        logger.info("Signed out")

    @api(version="2.6")
    def switch_site(self, site_item):
        url = "{0}/{1}".format(self.baseurl, "switchSite")
        switch_req = RequestFactory.Auth.switch_req(site_item.content_url)
        try:
            server_response = self.post_request(url, switch_req)
        except ServerResponseError as e:
            if e.code == "403070":
                return Auth.contextmgr(self.sign_out)
            else:
                raise e
        self.parent_srv._namespace.detect(server_response.content)
        self._check_status(server_response, url)
        parsed_response = fromstring(server_response.content)
        site_id = parsed_response.find(".//t:site", namespaces=self.parent_srv.namespace).get("id", None)
        user_id = parsed_response.find(".//t:user", namespaces=self.parent_srv.namespace).get("id", None)
        auth_token = parsed_response.find("t:credentials", namespaces=self.parent_srv.namespace).get("token", None)
        self.parent_srv._set_auth(site_id, user_id, auth_token)
        logger.info("Signed into {0} as user with id {1}".format(self.parent_srv.server_address, user_id))
        return Auth.contextmgr(self.sign_out)

    @api(version="3.10")
    def revoke_all_server_admin_tokens(self):
        url = "{0}/{1}".format(self.baseurl, "revokeAllServerAdminTokens")
        self.post_request(url, "")
        logger.info("Revoked all tokens for all server admins")



import copy
import logging
import warnings
from collections import namedtuple
from typing import TYPE_CHECKING, Callable, List, Optional, Tuple, Union

from .endpoint import Endpoint, api, parameter_added_in
from .exceptions import MissingRequiredFieldError
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import PaginationItem, ScheduleItem, TaskItem

logger = logging.getLogger("tableau.endpoint.schedules")
AddResponse = namedtuple("AddResponse", ("result", "error", "warnings", "task_created"))
OK = AddResponse(result=True, error=None, warnings=None, task_created=None)

if TYPE_CHECKING:
    from ..request_options import RequestOptions
    from ...models import DatasourceItem, WorkbookItem, FlowItem


class Schedules(Endpoint):
    @property
    def baseurl(self) -> str:
        return "{0}/schedules".format(self.parent_srv.baseurl)

    @property
    def siteurl(self) -> str:
        return "{0}/sites/{1}/schedules".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    @api(version="2.3")
    def get(self, req_options: Optional["RequestOptions"] = None) -> Tuple[List[ScheduleItem], PaginationItem]:
        logger.info("Querying all schedules")
        url = self.baseurl
        server_response = self.get_request(url, req_options)
        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_schedule_items = ScheduleItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_schedule_items, pagination_item

    @api(version="3.8")
    def get_by_id(self, schedule_id):
        if not schedule_id:
            error = "No Schedule ID provided"
            raise ValueError(error)
        logger.info("Querying a single schedule by id ({})".format(schedule_id))
        url = "{0}/{1}".format(self.baseurl, schedule_id)
        server_response = self.get_request(url)
        return ScheduleItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    @api(version="2.3")
    def delete(self, schedule_id: str) -> None:
        if not schedule_id:
            error = "Schedule ID undefined"
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, schedule_id)
        self.delete_request(url)
        logger.info("Deleted single schedule (ID: {0})".format(schedule_id))

    @api(version="2.3")
    def update(self, schedule_item: ScheduleItem) -> ScheduleItem:
        if not schedule_item.id:
            error = "Schedule item missing ID."
            raise MissingRequiredFieldError(error)

        url = "{0}/{1}".format(self.baseurl, schedule_item.id)
        update_req = RequestFactory.Schedule.update_req(schedule_item)
        server_response = self.put_request(url, update_req)
        logger.info("Updated schedule item (ID: {})".format(schedule_item.id))
        updated_schedule = copy.copy(schedule_item)
        return updated_schedule._parse_common_tags(server_response.content, self.parent_srv.namespace)

    @api(version="2.3")
    def create(self, schedule_item: ScheduleItem) -> ScheduleItem:
        if schedule_item.interval_item is None:
            error = "Interval item must be defined."
            raise MissingRequiredFieldError(error)

        url = self.baseurl
        create_req = RequestFactory.Schedule.create_req(schedule_item)
        server_response = self.post_request(url, create_req)
        new_schedule = ScheduleItem.from_response(server_response.content, self.parent_srv.namespace)[0]
        logger.info("Created new schedule (ID: {})".format(new_schedule.id))
        return new_schedule

    @api(version="2.8")
    @parameter_added_in(flow="3.3")
    def add_to_schedule(
        self,
        schedule_id: str,
        workbook: Optional["WorkbookItem"] = None,
        datasource: Optional["DatasourceItem"] = None,
        flow: Optional["FlowItem"] = None,
        task_type: Optional[str] = None,
    ) -> List[AddResponse]:
        # There doesn't seem to be a good reason to allow one item of each type?
        if workbook and datasource:
            warnings.warn("Passing in multiple items for add_to_schedule will be deprecated", PendingDeprecationWarning)
        items: List[
            Tuple[str, Union[WorkbookItem, FlowItem, DatasourceItem], str, Callable[[Optional[str], str], bytes], str]
        ] = []

        if workbook is not None:
            if not task_type:
                task_type = TaskItem.Type.ExtractRefresh
            items.append((schedule_id, workbook, "workbook", RequestFactory.Schedule.add_workbook_req, task_type))
        if datasource is not None:
            if not task_type:
                task_type = TaskItem.Type.ExtractRefresh
            items.append((schedule_id, datasource, "datasource", RequestFactory.Schedule.add_datasource_req, task_type))
        if flow is not None and not (workbook or datasource):  # Cannot pass a flow with any other type
            if not task_type:
                task_type = TaskItem.Type.RunFlow
            items.append(
                (schedule_id, flow, "flow", RequestFactory.Schedule.add_flow_req, task_type)
            )  # type:ignore[arg-type]

        results = (self._add_to(*x) for x in items)
        # list() is needed for python 3.x compatibility
        return list(filter(lambda x: not x.result, results))  # type:ignore[arg-type]

    def _add_to(
        self,
        schedule_id,
        resource: Union["DatasourceItem", "WorkbookItem", "FlowItem"],
        type_: str,
        req_factory: Callable[
            [
                str,
                str,
            ],
            bytes,
        ],
        item_task_type,
    ) -> AddResponse:
        id_ = resource.id
        url = "{0}/{1}/{2}s".format(self.siteurl, schedule_id, type_)
        add_req = req_factory(id_, task_type=item_task_type)  # type: ignore[call-arg, arg-type]
        response = self.put_request(url, add_req)

        error, warnings, task_created = ScheduleItem.parse_add_to_schedule_response(response, self.parent_srv.namespace)
        if task_created:
            logger.info("Added {} to {} to schedule {}".format(type_, id_, schedule_id))

        if error is not None or warnings is not None:
            return AddResponse(
                result=False,
                error=error,
                warnings=warnings,
                task_created=task_created,
            )
        else:
            return OK



class Sort(object):
    def __init__(self, field, direction):
        self.field = field
        self.direction = direction

    def __str__(self):
        return "{0}:{1}".format(self.field, self.direction)



from defusedxml.ElementTree import fromstring

from .property_decorators import property_not_empty


class ColumnItem(object):
    def __init__(self, name, description=None):
        self._id = None
        self.description = description
        self.name = name

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @name.setter
    @property_not_empty
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def remote_type(self):
        return self._remote_type

    def _set_values(self, id, name, description, remote_type):
        if id is not None:
            self._id = id
        if name:
            self._name = name
        if description:
            self.description = description
        if remote_type:
            self._remote_type = remote_type

    @classmethod
    def from_response(cls, resp, ns):
        all_column_items = list()
        parsed_response = fromstring(resp)
        all_column_xml = parsed_response.findall(".//t:column", namespaces=ns)

        for column_xml in all_column_xml:
            (id, name, description, remote_type) = cls._parse_element(column_xml, ns)
            column_item = cls(name)
            column_item._set_values(id, name, description, remote_type)
            all_column_items.append(column_item)

        return all_column_items

    @staticmethod
    def _parse_element(column_xml, ns):
        id = column_xml.get("id", None)
        name = column_xml.get("name", None)
        description = column_xml.get("description", None)
        remote_type = column_xml.get("remoteType", None)

        return id, name, description, remote_type



####
# This script demonstrates how to update a connections credentials on a server to embed the credentials
#
# To run the script, you must have installed Python 3.7 or later.
####

import argparse
import logging

import tableauserverclient as TSC


def main():
    parser = argparse.ArgumentParser(description="Update a connection on a datasource or workbook to embed credentials")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    parser.add_argument("resource_type", choices=["workbook", "datasource"])
    parser.add_argument("resource_id")
    parser.add_argument("connection_id")
    parser.add_argument("datasource_username")
    parser.add_argument("datasource_password")

    args = parser.parse_args()

    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        endpoint = {"workbook": server.workbooks, "datasource": server.datasources}.get(args.resource_type)

        update_function = endpoint.update_connection
        resource = endpoint.get_by_id(args.resource_id)
        endpoint.populate_connections(resource)
        connections = list(filter(lambda x: x.id == args.connection_id, resource.connections))
        assert len(connections) == 1
        connection = connections[0]
        connection.username = args.datasource_username
        connection.password = args.datasource_password
        connection.embed_password = True
        print(update_function(resource, connection).__dict__)


if __name__ == "__main__":
    main()



####
# This script demonstrates how to set the refresh schedule for
# a workbook or datasource.
#
# To run the script, you must have installed Python 3.7 or later.
####


import argparse
import logging

import tableauserverclient as TSC


def usage(args):
    parser = argparse.ArgumentParser(description="Set refresh schedule for a workbook or datasource.")
    # Common options; please keep those in sync across all samples
    parser.add_argument("--server", "-s", required=True, help="server address")
    parser.add_argument("--site", "-S", help="site name")
    parser.add_argument(
        "--token-name", "-p", required=True, help="name of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--token-value", "-v", required=True, help="value of the personal access token used to sign into the server"
    )
    parser.add_argument(
        "--logging-level",
        "-l",
        choices=["debug", "info", "error"],
        default="error",
        help="desired logging level (set to error by default)",
    )
    # Options specific to this sample
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--workbook", "-w")
    group.add_argument("--datasource", "-d")
    group.add_argument("--flow", "-f")
    parser.add_argument("schedule")

    return parser.parse_args(args)


def make_filter(**kwargs):
    options = TSC.RequestOptions()
    for item, value in kwargs.items():
        name = getattr(TSC.RequestOptions.Field, item)
        options.filter.add(TSC.Filter(name, TSC.RequestOptions.Operator.Equals, value))
    return options


def get_workbook_by_name(server, name):
    request_filter = make_filter(Name=name)
    workbooks, _ = server.workbooks.get(request_filter)
    assert len(workbooks) == 1
    return workbooks.pop()


def get_datasource_by_name(server, name):
    request_filter = make_filter(Name=name)
    datasources, _ = server.datasources.get(request_filter)
    assert len(datasources) == 1
    return datasources.pop()


def get_flow_by_name(server, name):
    request_filter = make_filter(Name=name)
    flows, _ = server.flows.get(request_filter)
    assert len(flows) == 1
    return flows.pop()


def get_schedule_by_name(server, name):
    schedules = [x for x in TSC.Pager(server.schedules) if x.name == name]
    assert len(schedules) == 1
    return schedules.pop()


def assign_to_schedule(server, workbook_or_datasource, schedule):
    server.schedules.add_to_schedule(schedule.id, workbook_or_datasource)


def run(args):
    # Set logging level based on user input, or error by default
    logging_level = getattr(logging, args.logging_level.upper())
    logging.basicConfig(level=logging_level)

    # Step 1: Sign in to server.
    tableau_auth = TSC.PersonalAccessTokenAuth(args.token_name, args.token_value, site_id=args.site)
    server = TSC.Server(args.server, use_server_version=True)
    with server.auth.sign_in(tableau_auth):
        if args.workbook:
            item = get_workbook_by_name(server, args.workbook)
        elif args.datasource:
            item = get_datasource_by_name(server, args.datasource)
        elif args.flow:
            item = get_flow_by_name(server, args.flow)
        else:
            print("A scheduleable item must be included")
            return
        schedule = get_schedule_by_name(server, args.schedule)

        assign_to_schedule(server, item, schedule)


def main():
    import sys

    args = usage(sys.argv[1:])
    run(args)


if __name__ == "__main__":
    main()



import logging

from .endpoint import Endpoint, api
from .exceptions import MissingRequiredFieldError
from tableauserverclient.server import RequestFactory
from tableauserverclient.models import SubscriptionItem, PaginationItem

logger = logging.getLogger("tableau.endpoint.subscriptions")

from typing import List, Optional, TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from ..request_options import RequestOptions


class Subscriptions(Endpoint):
    @property
    def baseurl(self) -> str:
        return "{0}/sites/{1}/subscriptions".format(self.parent_srv.baseurl, self.parent_srv.site_id)

    @api(version="2.3")
    def get(self, req_options: Optional["RequestOptions"] = None) -> Tuple[List[SubscriptionItem], PaginationItem]:
        logger.info("Querying all subscriptions for the site")
        url = self.baseurl
        server_response = self.get_request(url, req_options)

        pagination_item = PaginationItem.from_response(server_response.content, self.parent_srv.namespace)
        all_subscriptions = SubscriptionItem.from_response(server_response.content, self.parent_srv.namespace)
        return all_subscriptions, pagination_item

    @api(version="2.3")
    def get_by_id(self, subscription_id: str) -> SubscriptionItem:
        if not subscription_id:
            error = "No Subscription ID provided"
            raise ValueError(error)
        logger.info("Querying a single subscription by id ({})".format(subscription_id))
        url = "{}/{}".format(self.baseurl, subscription_id)
        server_response = self.get_request(url)
        return SubscriptionItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    @api(version="2.3")
    def create(self, subscription_item: SubscriptionItem) -> SubscriptionItem:
        if not subscription_item:
            error = "No Susbcription provided"
            raise ValueError(error)
        logger.info("Creating a subscription ({})".format(subscription_item))
        url = self.baseurl
        create_req = RequestFactory.Subscription.create_req(subscription_item)
        server_response = self.post_request(url, create_req)
        return SubscriptionItem.from_response(server_response.content, self.parent_srv.namespace)[0]

    @api(version="2.3")
    def delete(self, subscription_id: str) -> None:
        if not subscription_id:
            error = "Subscription ID undefined."
            raise ValueError(error)
        url = "{0}/{1}".format(self.baseurl, subscription_id)
        self.delete_request(url)
        logger.info("Deleted subscription (ID: {0})".format(subscription_id))

    @api(version="2.3")
    def update(self, subscription_item: SubscriptionItem) -> SubscriptionItem:
        if not subscription_item.id:
            error = "Subscription item missing ID. Subscription must be retrieved from server first."
            raise MissingRequiredFieldError(error)
        url = "{0}/{1}".format(self.baseurl, subscription_item.id)
        update_req = RequestFactory.Subscription.update_req(subscription_item)
        server_response = self.put_request(url, update_req)
        logger.info("Updated subscription item (ID: {0})".format(subscription_item.id))
        return SubscriptionItem.from_response(server_response.content, self.parent_srv.namespace)[0]
